/*
	libmatroska : parse and create Matroska files, see http://matroska.sf.net/
	Copyright (C) 2002 Steve Lhomme

	This library is free software; you can redistribute it and/or
	modify it under the terms of the GNU Lesser General Public
	License as published by the Free Software Foundation; either
	version 2.1 of the License, or (at your option) any later version.
	
	This library is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
	Lesser General Public License for more details.
	
	You should have received a copy of the GNU Lesser General Public
	License along with this library; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

/*!
	\file
	\version \$Id$
	\author Steve Lhomme     <robux4 @ users.sf.net>
*/

#include <cassert>

#include "StdInclude.h"
#include "EbmlMaster.h"

START_LIBEBML_NAMESPACE

EbmlMaster::EbmlMaster(const bool bSizeIsknown)
 :EbmlElement(0), Context(NULL), bSizeIsFinite(bSizeIsknown)
{}
	
EbmlMaster::EbmlMaster(const EbmlRawElement & ElementToClone)
	:EbmlElement(0), Context(NULL), bSizeIsFinite(true)
{
	*(static_cast<EbmlRawElement*>(this)) = ElementToClone;
}

/*!
	\todo handle exception on errors
	\todo write all the Mandatory elements in the Context, otherwise assert
*/
uint32 EbmlMaster::RenderData(IOCallback & output)
{
	uint32 Result = 0;
	std::list<EbmlElement *>::const_iterator ElementListIdx;
	
	for (ElementListIdx = ElementList.begin(); ElementListIdx != ElementList.end(); ElementListIdx++) {
		Result += (*ElementListIdx)->Render(output);
	}

	return Result;
}

/*!
	\todo We might be able to forbid elements that don't exist in the context
*/
bool EbmlMaster::PushElement(EbmlElement & element)
{
	ElementList.push_back(&element);
	return true;
}

uint64 EbmlMaster::UpdateSize()
{
	Size = 0;

	if (!bSizeIsFinite)
		return (0-1);

	std::list<EbmlElement *>::const_iterator ElementListIdx;
	
	for (ElementListIdx = ElementList.begin(); ElementListIdx != ElementList.end(); ElementListIdx++) {
		(*ElementListIdx)->UpdateSize();
		uint64 SizeToAdd = (*ElementListIdx)->ElementSize();
		if (SizeToAdd == (0-1))
			return (0-1);
		Size += SizeToAdd;
	}

	return Size;
}

uint64 EbmlMaster::ReadData(IOCallback & input)
{
	input.setFilePointer(Size, seek_current);
	return Size;
}

bool EbmlMaster::ProcessMandatory()
{
	if (Context == NULL || Context->Size == 0)
	{
		return true;
	}

	assert(Context->MyTable != NULL);

	unsigned int EltIdx;
	for (EltIdx = 0; EltIdx < Context->Size; EltIdx++) {
		if (Context->MyTable[EltIdx].Mandatory && Context->MyTable[EltIdx].Unique) {
			assert(Context->MyTable[EltIdx].GetCallbacks->Create != NULL);
			PushElement(Context->MyTable[EltIdx].GetCallbacks->Create());
		}
	}
	
	return true;
}

EbmlElement *EbmlMaster::FindFirstElt(const EbmlCallbacks & Callbacks, const bool bCreateIfNull)
{
	std::list<EbmlElement *>::const_iterator ElementListIdx;
	for (ElementListIdx = ElementList.begin(); ElementListIdx != ElementList.end(); ElementListIdx++) {
		if (EbmlId(*(*ElementListIdx)) == Callbacks.GlobalId)
			return *ElementListIdx;
	}
	
	if (bCreateIfNull && Callbacks.Create != NULL) {
		// add the element
		EbmlElement *NewElt = &(Callbacks.Create());
		if (NewElt == NULL)
			return NULL;

		if (!PushElement(*NewElt)) {
			delete NewElt;
			NewElt = NULL;
		}
		return NewElt;
	}
	
	return NULL;
}

EbmlElement *EbmlMaster::AddNewElt(const EbmlCallbacks & Callbacks)
{
	// add the element
	EbmlElement *NewElt = &(Callbacks.Create());
	if (NewElt == NULL)
		return NULL;

	if (!PushElement(*NewElt)) {
		delete NewElt;
		NewElt = NULL;
	}
	return NewElt;
}

END_LIBEBML_NAMESPACE
