/*
	libmatroska : parse and create Matroska files, see http://matroska.sf.net/
	Copyright (C) 2002 Steve Lhomme

	This library is free software; you can redistribute it and/or
	modify it under the terms of the GNU Lesser General Public
	License as published by the Free Software Foundation; either
	version 2.1 of the License, or (at your option) any later version.
	
	This library is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
	Lesser General Public License for more details.
	
	You should have received a copy of the GNU Lesser General Public
	License along with this library; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

/*!
	\file
	\version \$Id$
	\author Steve Lhomme     <robux4 @ users.sf.net>
*/

#include <cassert>

#include "StdInclude.h"
#include "EbmlRawElement.h"
#include "EbmlElement.h"

START_LIBEBML_NAMESPACE

class DummyRawElement : public EbmlRawElement {
	public:
		DummyRawElement(const uint64 DefaultSize) :EbmlRawElement(DefaultSize) {}
		operator const EbmlId &() const {return DummyId;}
	protected:
		static EbmlId DummyId;
};
EbmlId DummyRawElement::DummyId = {{0xFF}, 1};

EbmlRawElement::EbmlRawElement(const uint64 DefaultSize)
 :SizeLength(0) ///< write optimal size by default
{
	Size = DefaultSize;
	bSizeOptimal = (DefaultSize == 0);
}

/*!
	\todo handle exception on errors
	\todo better sanity check (more than the first octet tested)
* /
bool EbmlRawElement::SetID(const EbmlId & NewId)
{
	// sanity checks
	if (NewId.Length <= 0)
		return false;
	
	// create the mask to verify the ID match the Class scheme
	int IdMask = 1 << 7;
	int IdScheme = 1 << 7;
	unsigned int i;
	for (i=1; i<NewId.Length; i++) {
		IdMask = IdMask>>1 | 1<<7;
		IdScheme >>= 1;
	}
	
	if ((NewId.Value[0] & IdMask) != IdScheme)
		return false;
	
	// ID is OK
	MyId = NewId;

	return true;
}*/

/*!
	\todo handle more than CodedSize of 5
*/
int EbmlRawElement::CodedSizeLength() const
{
	int CodedSize;
	// prepare the head of the size (000...01xxxxxx)
	// optimal size
	if (Size < 127) // 2^7 - 1
		CodedSize = 1;
	else if (Size < 16383) // 2^14 - 1
		CodedSize = 2;
	else if (Size < 2097151L) // 2^21 - 1
		CodedSize = 3;
	else if (Size < 268435455L) // 2^28 - 1
		CodedSize = 4;
	else CodedSize = 5;

	if (SizeLength > 0 && CodedSize < SizeLength) {
		// defined size
		CodedSize = SizeLength;
	}

	return CodedSize;
}

/*
uint64 EbmlRawElement::ElementSize() const
{
	return Size + EbmlId(*this).Length + CodedSizeLength();
}
*/

/*!
	\todo handle errors (skip first byte and try again until a valid is found or no data)
	\todo replace the new RawElement with the appropriate class (when known)
*/
EbmlRawElement * EbmlRawElement::FindNextID(IOCallback & DataStream, const struct EbmlSemanticContext & Context, const uint64 MaxDataSize, const bool AllowUnknownSize)
{
	EbmlId PossibleID;
	PossibleID.Length = 0;
	binary PossibleSize[8]; // we don't support size stored in more than 64 bits
	unsigned int PossibleSizeLength = 0;
	
	assert(AllowUnknownSize != true); // Not supported yet (requires a context)
	
	// read ID
	uint32 ReadSize = 0;
	binary BitMask = 1 << 7;
	while (1) {
		ReadSize += DataStream.read(&PossibleID.Value[PossibleID.Length], 1);
		if (ReadSize == PossibleID.Length) {
			return NULL; // no more data ?
		}
		if (++PossibleID.Length > 4) {
			return NULL; // we don't support element IDs over class D
		}
		if (PossibleID.Value[0] & BitMask) {
			// this is the last element of the ID
			break;
		}
		BitMask >>= 1;
	}
	
	// read the data size
	BitMask = 1 << 7;
	while (1) {
		ReadSize += DataStream.read(&PossibleSize[PossibleSizeLength], 1);
		if (ReadSize == PossibleSizeLength) {
			return NULL; // no more data ?
		}
		if (++PossibleSizeLength > 8) {
			return NULL; // we don't support element IDs over class D
		}
		if (PossibleSize[0] & BitMask) {
			// this is the last element of the ID
			break;
		}
		BitMask >>= 1;
	}
	
	/// \todo find the element in the context
	/// \todo set the correct ID to this element
	EbmlRawElement *Result = new DummyRawElement(0);
	
////	Result->SetID(PossibleID);
	Result->SetSizeLength(PossibleSizeLength);
	
	Result->Size = 0;
	Result->Size |= PossibleSize[0] & ~BitMask;
	for (unsigned int i = 1; i<PossibleSizeLength; i++) {
		Result->Size <<= 8;
		Result->Size |= PossibleSize[i];
	}
	
	return Result;
}

/*!
	\todo replace the new RawElement with the appropriate class (when known)
*/
EbmlRawElement * EbmlRawElement::FindNextID(IOCallback & DataStream, const struct EbmlSemanticContext & Context, const EbmlCallbacks & ClassInfos, const uint64 MaxDataSizeconst, const bool AllowUnknownSize)
{
	EbmlId PossibleID;
	PossibleID.Length = 0;
	binary PossibleSize[8]; // we don't support size stored in more than 64 bits
	unsigned int PossibleSizeLength = 0;
	bool bElementFound = false;
	
	assert(AllowUnknownSize != true); // Not supported yet (requires a context)
	
	binary BitMask;
	while (!bElementFound) {
		// read ID
		uint32 ReadSize = 0;
		BitMask = 1 << 7;
		while (1) {
			ReadSize += DataStream.read(&PossibleID.Value[PossibleID.Length], 1);
			if (ReadSize == PossibleID.Length) {
				return NULL; // no more data ?
			}
			if (++PossibleID.Length > 4) {
				return NULL; // we don't support element IDs over class D
			}
			if (PossibleID.Value[0] & BitMask) {
				// this is the last element of the ID
				// check wether that's the one we're looking for
/*			if (PossibleID == ClassInfos.GlobalId) {
					break;
				} else {
					/// \todo This element should be skipped (use a context ?)
				}*/
				bElementFound = true; /// \todo not exactly the one we're looking for
				break;
			}
			BitMask >>= 1;
		}
		
		// read the data size
		BitMask = 1 << 7;
		while (1) {
			ReadSize += DataStream.read(&PossibleSize[PossibleSizeLength], 1);
			if (ReadSize == PossibleSizeLength) {
				return NULL; // no more data ?
			}
			if (++PossibleSizeLength > 8) {
				return NULL; // we don't support element IDs over class D
			}
			if (PossibleSize[0] & BitMask) {
				// this is the last element of the ID
				break;
			}
			BitMask >>= 1;
		}
	}
	
	EbmlRawElement *Result = NULL;
	if (PossibleID == ClassInfos.GlobalId) {
		Result = static_cast<EbmlRawElement *>(&ClassInfos.Create());
	} else {
		/// \todo find the element in the context
		/// \todo set the correct ID to this element
		Result = new DummyRawElement(0);
	}
	
////	Result->SetID(PossibleID);
	Result->SetSizeLength(PossibleSizeLength);
	
	Result->Size = 0;
	Result->Size |= PossibleSize[0] & ~BitMask;
	for (unsigned int i = 1; i<PossibleSizeLength; i++) {
		Result->Size <<= 8;
		Result->Size |= PossibleSize[i];
	}
	
	return Result;
}

void EbmlRawElement::SkipData(IOCallback & DataStream)
{
	DataStream.setFilePointer(Size, seek_current);
}

END_LIBEBML_NAMESPACE
