/*
	libmatroska : parse and create Matroska files, see http://matroska.sf.net/
	Copyright (C) 2002 Steve Lhomme

	This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

/*!
	\file
	\version \$Id: ActualHeader.cpp,v 1.17 2002/07/14 20:41:41 robux4 Exp $
	\author Ingo Ralf Blum   <ingoralfblum @ users.sf.net>
	\author Lasse Krkkinen <tronic @ users.sf.net>
	\author Steve Lhomme     <robux4 @ users.sf.net>
*/
#include "StdInclude.hpp"
#include "ActualHeader.hpp"
#include "Adler32.hpp"
#include "Block.hpp"
#include "Cluster.hpp"
#include "SeekEntry.hpp"
#include "Track.hpp"
#include "AttachEntry.hpp"

START_MATROSKA_NAMESPACE

const uint8 ActualHeader::myVersion = 0; // update when we can red newer version of Matroska
const SmallElement ActualHeader::myTrackElement     = {0x00,0x00,0x00,0x01};
const SmallElement ActualHeader::myCodecElement     = {0x00,0x00,0x00,0x02};
const SmallElement ActualHeader::mySeekElement      = {0x00,0x00,0x00,0x0E};
const SmallElement ActualHeader::myChapterElement   = {0x00,0x00,0x00,0x08};
const SmallElement ActualHeader::myAttachElement    = {0x00,0x00,0x00,0x09};
const SmallElement ActualHeader::mySignatureElement = {0x00,0x00,0x00,0x0C};

/*!
    \todo handle data properly
*/
uint32 ActualHeader::Render(IOCallback & output, const StreamInfo & aStreamInfo) const
{
	uint32 written_size = default_size();
	
	try {
		int i;
	binary *Header = new binary[written_size];
	memset(Header, 0x00, written_size);

	binary *cursor = Header;

	// versions
	*cursor++ = myVersion;
	*cursor++ = myMinVersion; // min version to use all
	*cursor++ = myAbsMinVersion; // min version to read

	// flags
	*cursor++ = (myLinearWriteMode?0x80:0x00);

	// time length
	big_uint32 bu32 = aStreamInfo.TimeLength;
	bu32.Fill(cursor);
	cursor += sizeof(uint32);

	// main header length
	bu32 = aStreamInfo.MainHeaderSize;
	bu32.Fill(cursor);
	cursor += sizeof(uint32);

	// Cluster Positions
	cursor = &Header[0x10];
	big_uint64 bu64 = aStreamInfo.ClusterPosition;
	bu64.Fill(cursor);
	cursor += sizeof(uint64);

	bu64 = aStreamInfo.ClusterSize;
	bu64.Fill(cursor);
	cursor += sizeof(uint64);

	cursor = &Header[0x60];

	// Track Entries position
	bu64 = aStreamInfo.TrackEntryPosition;
	bu64.Fill(cursor);
	cursor += sizeof(uint64);

	bu32 = aStreamInfo.TrackEntriesSize;
	bu32.Fill(cursor);
	cursor += sizeof(uint32);

	for (i=0;i<4;i++)
	    *cursor++ = myTrackElement[i];

	// Codec Entries position
	bu64 = aStreamInfo.CodecEntryPosition;
	bu64.Fill(cursor);
	cursor += sizeof(uint64);

	bu32 = aStreamInfo.CodecEntrySize;
	bu32.Fill(cursor);
	cursor += sizeof(uint32);

	for (i=0;i<4;i++)
	    *cursor++ = myCodecElement[i];

	// Seek Entries position
	bu64 = aStreamInfo.SeekEntryPosition;
	bu64.Fill(cursor);
	cursor += sizeof(uint64);

	bu32 = aStreamInfo.SeekEntriesSize;
	bu32.Fill(cursor);
	cursor += sizeof(uint32);

	for (i=0;i<4;i++)
	    *cursor++ = mySeekElement[i];

	// Chapter Entries position
	bu64 = aStreamInfo.ChapterEntryPosition;
	bu64.Fill(cursor);
	cursor += sizeof(uint64);

	bu32 = aStreamInfo.ChapterEntrySize;
	bu32.Fill(cursor);
	cursor += sizeof(uint32);

	for (i=0;i<4;i++)
	    *cursor++ = myChapterElement[i];

	// Attach Entries position
	bu64 = aStreamInfo.AttachEntryPosition;
	bu64.Fill(cursor);
	cursor += sizeof(uint64);

	bu32 = aStreamInfo.AttachEntrySize;
	bu32.Fill(cursor);
	cursor += sizeof(uint32);

	for (i=0;i<4;i++)
	    *cursor++ = myAttachElement[i];

	// Signature Entries position
	bu64 = aStreamInfo.SignatureEntryPosition;
	bu64.Fill(cursor);
	cursor += sizeof(uint64);

	bu32 = aStreamInfo.SignatureEntrySize;
	bu32.Fill(cursor);
	cursor += sizeof(uint32);

	for (i=0;i<4;i++)
	    *cursor++ = mySignatureElement[i];

	cursor = &Header[0x160];
	// Filename
	memcpy(cursor, myOriginalFilename.buffer(), myOriginalFilename.size());
	cursor += myOriginalFilename.size();

	// Next part
	memcpy(cursor, myNextPartFilename.buffer(), myNextPartFilename.size());
	cursor += myOriginalFilename.size();
	bu32 = myNextPartTimecode;
	bu32.Fill(cursor);
	cursor += sizeof(uint32);

	// Encoding data
	memcpy(cursor, aStreamInfo.EncoderLib.buffer(), aStreamInfo.EncoderLib.size());
	cursor += aStreamInfo.EncoderLib.size();

	memcpy(cursor, aStreamInfo.EncoderApp.buffer(), aStreamInfo.EncoderApp.size());
	cursor += aStreamInfo.EncoderApp.size();

	cursor = &Header[0x260];
	// number of blocks
	bu32 = aStreamInfo.NumberBlock;
	bu32.Fill(cursor);
	cursor += sizeof(uint32);

	*cursor++ = CLUSTER_HEADER_SIZE;
	*cursor++ = BLOCK_HEADER_SIZE;
	*cursor++ = CLUSTER_TRAILER_SIZE;

	*cursor++ = SEEK_HEADER_SIZE;

	/// \todo Checksum of seek entries
	cursor += sizeof(uint32);

	// track entry size
	big_uint16 bu16 = Track::default_size();
	bu16.Fill(cursor);
	cursor += sizeof(uint16);

	/// \todo  subtrack entry size
	cursor += sizeof(uint16);

	/// \todo Checksum of attach entries
	cursor += sizeof(uint32);

	// attach entry size
	bu16 = ATTACH_ENTRY_SIZE;
	bu16.Fill(cursor);
	cursor += sizeof(uint16);

	cursor++;
	/// \todo Number of chapter editions
	/// \todo Checksum of chapter eidtions
	/// \todo Size of chapter block
	/// \todo Size of edition entry

	// Checksum
	Adler32 checksum;
	checksum.Feed(Header,written_size - 4);
	memcpy(&Header[written_size-4],&checksum.endian(),4);

	output.writeFully(Header,written_size);

    }
    catch (exception & Ex)
    {
	throw Ex;
    }

    return written_size;
}

uint32 ActualHeader::Read(IOCallback & input, StreamInfo & aStreamInfo)
{
    try {
	binary Header[864];
	input.readFully(Header, default_size());
	uint32 result = default_size();

	if (result != default_size())
	    throw 0; // unexpected end of file

	// verify checksum
	Adler32 checksum;
	checksum.Feed(Header, default_size() - 4);
	if (checksum != &Header[default_size()-4])
	    throw 0;

	// check version
	binary version = Header[0];
	if (version > myVersion)
	{
    	    version = Header[1];
	    if (version > myVersion)
	    {
    		version = Header[2];
		if (version > myVersion)
		{
		    // we cannot read this file
    		    myUseMode = USE_NONE;
		    throw 0;
		}
		else
		{
		    // we can only read the file
    		    myUseMode = USE_READ;
		}
	    }
	    else
	    {
		// we can use all features here
		myUseMode = USE_ALL;
	    }
	}

	binary *cursor = &Header[3];
	// process all for reading version
	myLinearWriteMode = (*cursor++ & 0x01) != 0;

	// time length
	big_uint32 bu32;
	int i;
	bu32.Eval(cursor);
	cursor += sizeof(uint32);
	aStreamInfo.TimeLength = bu32;

	/// \todo Handle Size of Main Header (uint32)
	bu32.Eval(cursor);
	uint32 MainHeadSize = bu32;
	cursor += sizeof(uint32);
	if (aStreamInfo.MainHeaderSize != MainHeadSize)
	    throw 0; // can't read that

	// Cluster position
	cursor = &Header[0x10];
	big_uint64 bu64;
	bu64.Eval(cursor);
	cursor += sizeof(uint64);
	aStreamInfo.ClusterPosition = bu64;

	bu64.Eval(cursor);
	cursor += sizeof(uint64);
	aStreamInfo.ClusterSize = bu64;

	// Track position
	cursor = &Header[0x60];
	bu64.Eval(cursor);
	cursor += sizeof(uint64);
	aStreamInfo.TrackEntryPosition = bu64;

	bu32.Eval(cursor);
	cursor += sizeof(uint32);
	aStreamInfo.TrackEntriesSize = bu32;

	for (i=0; i<4; i ++){
	    if (myTrackElement[i] != *cursor++)
		throw 0; // this is not the Track Entry !
	}

	// Codec Header position
	bu64.Eval(cursor);
	cursor += sizeof(uint64);
	aStreamInfo.CodecEntryPosition = bu64;

	bu32.Eval(cursor);
	cursor += sizeof(uint32);
	aStreamInfo.CodecEntrySize = bu32;

	for (i=0; i<4; i ++){
	    if (myCodecElement[i] != *cursor++)
		throw 0; // this is not the Codec Entry !
	}

	// Filename
	cursor = &Header[0x160];
	myOriginalFilename = (char*) cursor;
	cursor += myOriginalFilename.size();

	// Next part timecode
	myNextPartFilename = (char*) cursor;
	cursor += myNextPartFilename.size();

	bu32.Eval(cursor);
	cursor += sizeof(uint32);
	myNextPartTimecode = bu32;

	// Encoder data
	aStreamInfo.EncoderLib = (char*) cursor;
	cursor += aStreamInfo.EncoderLib.size();

	aStreamInfo.EncoderApp = (char*) cursor;
	cursor += aStreamInfo.EncoderApp.size();

	// number of blocks
	cursor = &Header[0x260];
	bu32.Eval(cursor);
	cursor += sizeof(uint32);
	aStreamInfo.NumberBlock = bu32;

	// Small element sizes
	aStreamInfo.ClusterHeadSize = *cursor++;
	if (aStreamInfo.ClusterHeadSize < CLUSTER_HEADER_SIZE)
	    throw 0; // impossible to read that

	aStreamInfo.BlockHeadSize = *cursor++;
	if (aStreamInfo.BlockHeadSize < BLOCK_HEADER_SIZE)
	    throw 0; // impossible to read that

	aStreamInfo.ClusterFootSize = *cursor++;
	if (aStreamInfo.ClusterFootSize < CLUSTER_TRAILER_SIZE)
	    throw 0; // impossible to read that

	aStreamInfo.SeekEntrySize = *cursor++;
	if (aStreamInfo.SeekEntrySize < SEEK_HEADER_SIZE)
	    throw 0; // impossible to read that

	// Seek Entries Checksum
	cursor += sizeof(uint32);

	big_uint16 bu16;
	bu16.Eval(cursor);
	cursor += sizeof(uint16);
	aStreamInfo.TrackEntrySize = bu16;

	if (myUseMode == USE_READ)
	    return result;
	
	/// \todo process all for full access version

	return result;
    }
    catch (exception & Ex)
    {
	throw Ex;
    }
}

END_MATROSKA_NAMESPACE
