/*
	libmatroska : parse and create Matroska files, see http://matroska.sf.net/
	Copyright (C) 2002 Steve Lhomme

	This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

/*!
	\file
	\version \$Id: BlockFxS.cpp,v 1.21 2002/11/20 22:37:39 robux4 Exp $
	\author Lasse Krkkinen <tronic @ users.sf.net>
	\author Steve Lhomme     <robux4 @ users.sf.net>
*/
#include "StdInclude.hpp"
#include "Frame.hpp"
#include "BlockFxS.hpp"

#include <sstream>

using namespace std;

START_MATROSKA_NAMESPACE

BlockFxS::BlockFxS(const uint8 track_number, const uint32 timecode, const uint32 frame_size)
 :Block(track_number,timecode)
 ,FrameSize(frame_size)
 ,myFrame(NULL)
{
    if (frame_size==0)
    {
	stringstream Msg;
	Msg << "Block created with null frame size";
	throw runtime_error(Msg.str());
    }
}

/*!
    \todo add a verification between frame_size and FrameSize
*/
BlockFxS::BlockFxS(const uint8 track_number, const uint32 frame_size, const binary * aFrame, const uint32 aFrameSize)
 :Block(track_number, aFrame, aFrameSize)
 ,FrameSize(frame_size)
 ,myFrame(NULL)
{
    aFrame += BLOCK_HEADER_SIZE;
    AddFrame(aFrame, aFrameSize - BLOCK_HEADER_SIZE, KeyFrame, BFrame);
}

/*!
    \return wether we should render the current block or not
*/
bool BlockFxS::AddFrame(const binary * buffer, const uint32 buffer_size, const bool a_KeyFrame, const bool a_BFrame)
{
    if (buffer == NULL)
    {
	stringstream Msg;
	Msg << "The buffer you want to write is NULL";
	throw runtime_error(Msg.str());
    }
    
    if (buffer_size != FrameSize)
    {
	stringstream Msg;
	Msg << "Data length doesn't correspond to fixed block length";
	throw runtime_error(Msg.str()); // we only allow frames of the same length
    }
    
    if (myFrame != NULL)
    {
	stringstream Msg;
	Msg << "Block is full";
	throw runtime_error(Msg.str());
    }

    // here we know that this frame can fit in this block
    try {
	myFrame = new Frame(buffer, buffer_size);
	myFrameRead = false;
    }
    catch(string & Ex)
    {
	throw Ex;
    }
    catch (exception & Ex)
    {
	throw Ex; // I hope it catches the
    }
    
    KeyFrame = a_KeyFrame;
    BFrame = a_BFrame;

    return true; // only one frame per Block
}

/*!
    \todo return the number of bytes written
 */
uint32 BlockFxS::Render(IOCallback & output, Adler32 & checksum)
{
    if (myFrame == NULL)
    {
	stringstream Msg;
	Msg << "No data in Block, NULL fixed size Block not allowed";
	throw runtime_error(Msg.str()); // we only allow frames of the same length
    }

    // write header
    uint32 result = RenderHeader(output, checksum);

    // write buffer
    output.writeFully(myFrame->buf(),FrameSize);
    checksum.Feed(myFrame->buf(),FrameSize);

	// free the frame
	delete myFrame;
	myFrame = NULL;

    return result + FrameSize;
}

bool BlockFxS::GetNextFrame(uint32 & aTimecode, const Frame * & aFrame,
			      bool & aKeyFrame, bool & aBFrame)
{
    if (myFrame == NULL || myFrameRead)
	return false;

    aFrame = myFrame;
	aTimecode = TimeCode;
	aKeyFrame = KeyFrame;
	aBFrame = BFrame;

    myFrameRead = true;

    return true;
}

END_MATROSKA_NAMESPACE
