/*
	libmatroska : parse and create Matroska files, see http://matroska.sf.net/
	Copyright (C) 2002 Steve Lhomme

	This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

/*!
    \file
    \version \$Id: BlockOcL.cpp,v 1.21 2002/11/20 22:37:39 robux4 Exp $
    \author Lasse Krkkinen <tronic @ users.sf.net>
    \author Steve Lhomme     <robux4 @ users.sf.net>
*/
#include "StdInclude.hpp"
#include "Frame.hpp"
#include "BlockOcL.hpp"

#include <sstream>

using namespace std;

START_MATROSKA_NAMESPACE

const uint16 BlockOcL::MAX_FRAME_SIZE_IN_LACE = 1200;
const uint8  BlockOcL::MAX_LACES_PER_FRAME = 5;

/*!
    \todo get the block header size for the upper part
*/
BlockOcL::BlockOcL(const uint8 track_number, const binary * aFrame, const uint32 aFrameSize)
 :Block(track_number, aFrame, aFrameSize)
{
    uint32 TotalBlockSize = aFrameSize;
    aFrame += BLOCK_HEADER_SIZE;
    uint8 LaceNb = *aFrame++; 
    uint32 lace_size;
    TotalBlockSize -= BLOCK_HEADER_SIZE + 1;
    while (LaceNb != 1)
    {
        // analyze the lace and create the corresponding frame
	lace_size = 0;
	while (*aFrame == 255)
	{
	    lace_size += 255;
	    aFrame++;
	}
	lace_size += *aFrame++;
	AddFrame(aFrame, lace_size, KeyFrame, BFrame);
	TotalBlockSize -= lace_size; // size of the found lace
	LaceNb--;
    }
    // this is the last frame
    AddFrame(aFrame, TotalBlockSize, KeyFrame, BFrame);

    FrameListIdx = FrameList.begin();
}

/*!
    \return wether we should render the current block or not
*/
bool BlockOcL::AddFrame(const binary * buffer, const uint32 buffer_size, const bool a_KeyFrame, const bool a_BFrame)
{
    if (buffer_size == 0)
	return true; // silently discard blank frames
    
    if (buffer == NULL)
    {
	stringstream Msg;
	Msg << "The buffer you want to write is NULL";
	throw runtime_error(Msg.str());
    }
    
    if (FrameList.size() >= MAX_LACES_PER_FRAME)
    {
	stringstream Msg;
	Msg << "Block is full, no more possible laces";
	throw runtime_error(Msg.str());// we already have enough laces
    }

    if (FrameList.size() != 0)
    {
	// the last block is to big to fit in a lace, it MUST be the last one
	if (FrameList.back()->length() >= MAX_FRAME_SIZE_IN_LACE)
	{
	    stringstream Msg;
	    Msg << "Block is full, last frame is too big";
	    throw runtime_error(Msg.str());
	}

	// verify that the frame is the same type as the block
	if (KeyFrame != a_KeyFrame || BFrame != a_BFrame)
	{
	    stringstream Msg;
	    Msg << "This frame cannot be saved in this block";
	    throw runtime_error(Msg.str());
	}
    }
    
    // here we know that this frame can fit in this block
    Frame * newFrame;
    try {
	newFrame = new Frame(buffer, buffer_size);
    }
    catch (exception & Ex)
    {
	stringstream Msg;
	Msg << "This frame cannot be created" << Ex.what();
	throw runtime_error(Msg.str());
    }
    
    KeyFrame = a_KeyFrame;
    BFrame = a_BFrame;

    FrameList.push_back(newFrame);
    
    if ((FrameList.size() >= 1 && size() >= MAX_FRAME_SIZE_IN_LACE) ||
	(FrameList.size() >= MAX_LACES_PER_FRAME))
	// this is the last frame that the block can hold
	return true;
    else
	// more frames can be added to the block
	return false;
}

/// \note we have to delete all the frames we allocated manually because we can't use auto_ptr
BlockOcL::~BlockOcL()
{
    for(size_t i=0; i<FrameList.size(); i++)
	delete FrameList[i];
}

/*!
    \todo return the number of bytes written
 */
uint32 BlockOcL::Render(IOCallback & output, Adler32 & checksum)
{
    if (FrameList.size() == 0)
    {
	stringstream Msg;
	Msg << "No data in Block, NULL laces not allowed";
	throw runtime_error(Msg.str()); // we only allow frames of the same length
    }

    // write header
    uint32 result = RenderHeader(output, checksum);

    uint32 data_size;
    uint32 lace_size;
    all_size(data_size, lace_size);

    // write the lace header to the file
    binary *lace_head,*lh_pointer;
    lace_head = new binary[lace_size]; /// \todo throw ?
    lh_pointer = lace_head;
    *lh_pointer++ = uint8(FrameList.size());
    uint32 frame_len;
    for(size_t frame_idx=0; frame_idx<FrameList.size()-1; frame_idx++) {
	frame_len = FrameList[frame_idx]->length();
	do {
	    if ((frame_len / 255) != 0)
	    {
		*lh_pointer++ = 255;
		frame_len -= 255;
	    }
	    if (frame_len < 255)
	    {
		*lh_pointer++ = frame_len;
		break;
	    }
	} while (1);
    }
    output.writeFully(lace_head,lace_size);
    checksum.Feed(lace_head,lace_size);

	result = size();

    // write the data to the file
    for(frame_idx=0; frame_idx<FrameList.size(); frame_idx++) {
	output.writeFully(FrameList[frame_idx]->buf(),FrameList[frame_idx]->length());
	checksum.Feed(FrameList[frame_idx]->buf(),FrameList[frame_idx]->length());

	//free the frame
	delete FrameList[frame_idx];
    }
	//clear the list
	FrameList.clear();

    return result;
}

inline uint32 BlockOcL::size() const
{
    uint32 data_size;
    uint32 lace_size;

    return all_size(data_size, lace_size);
}

inline uint32 BlockOcL::all_size(uint32 & data_size, uint32 & lace_size) const
{
    data_size = 0;
    lace_size = 0;

    size_t frame_idx;
    for(frame_idx=0; frame_idx<FrameList.size()-1; frame_idx++) {
	// 0 -> 0
	// 1 -> 1
	// ...
	// 254 -> 254
	// 255 -> 255,0
	// 256 -> 255,1
	// 509 -> 255,254
	// 510 -> 255,255,0
	// 511 -> 255,255,1
	// ...
	lace_size += (FrameList[frame_idx]->length() / 255) + 1;
	data_size += FrameList[frame_idx]->length();
    }
    data_size += FrameList.back()->length();
    lace_size++; // for the number of laces

    return BLOCK_HEADER_SIZE + data_size + lace_size;
}

bool BlockOcL::GetNextFrame(uint32 & aTimecode, const Frame * & aFrame,
			      bool & aKeyFrame, bool & aBFrame)
{
    if (FrameListIdx == FrameList.end())
	return false;

    aFrame = *FrameListIdx;
	aTimecode = TimeCode;
	aKeyFrame = KeyFrame;
	aBFrame = BFrame;

    FrameListIdx++;

    return true;
}

END_MATROSKA_NAMESPACE
