/*
	libmatroska : parse and create Matroska files, see http://matroska.sf.net/
	Copyright (C) 2002 Steve Lhomme

	This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

/*!
	\file
	\version \$Id: Cluster.cpp,v 1.22 2002/11/20 22:37:39 robux4 Exp $
	\warning The cluster used here are not exactly the same as in the current spec : There is no Number of the first Block and the Checksum is in the data trailer
	\author Ingo Ralf Blum   <ingoralfblum @ users.sf.net>
	\author Lasse Krkkinen <tronic @ users.sf.net>
	\author Steve Lhomme     <robux4 @ users.sf.net>
*/
#include "StdInclude.hpp"
#include "Cluster.hpp"
#include "Block.hpp"

#include <sstream>

using namespace std;

START_MATROSKA_NAMESPACE

const character Cluster::HEADER[4] = { 'C', 'H', 'd', 'r' };

bool Cluster::AddBlock(Block * a_Block)
{
    if (size() + a_Block->size() > MAX_SIZE)
	return false;

    myBlockList.push_back(a_Block);

    return true;
}

Cluster::~Cluster()
{
    for(size_t i=0; i<myBlockList.size(); i++) /// \todo might be a faster way
	delete myBlockList[i];
    if (myBuffer)
	delete myBuffer;
}

/*
    \todo Put the header and the checksum
    The header looks like this :
    0x00  "CHdr"
    0x04  Size of the cluster including header and trailer (uint32)
    0x08  Exact location of the Cluster in the file (uint64)
    The footer looks like this :
    0x00  Adler-32 checksum of Cluster Header and Block(s) of the Cluster
*/
uint32 Cluster::Render(IOCallback & output, uint32 & aNumberBlock) const
{
    binary Header[CLUSTER_HEADER_SIZE];
    binary *cursor = Header;
    *cursor++ = HEADER[0];
    *cursor++ = HEADER[1];
    *cursor++ = HEADER[2];
    *cursor++ = HEADER[3];
    
    uint32 size_written = size();
    big_uint32 cluster_size = size_written;
    cluster_size.Fill(cursor);
    cursor += sizeof(uint32);

    big_uint64 location;
    location = output.getFilePointer();
    location.Fill(cursor);
    cursor += sizeof(uint64);

    // write header
    output.writeFully(Header,CLUSTER_HEADER_SIZE);
    
    // compute checksum on the header
    Adler32 checksum;
    checksum.Feed(Header,CLUSTER_HEADER_SIZE);

    aNumberBlock = myBlockList.size();

    for(size_t i=0; i<myBlockList.size(); i++)
	myBlockList[i]->Render(output, checksum);

    // write trailer
    output.writeFully(&checksum.endian(),CLUSTER_TRAILER_SIZE);
    
    return size_written;
}

inline uint32 Cluster::size() const
{
    uint32 value = 0;
    value += CLUSTER_HEADER_SIZE;

    for(size_t i=0; i<myBlockList.size(); i++) /// \todo might be a faster way
	value += myBlockList[i]->size();

    value += CLUSTER_TRAILER_SIZE;

    return value;
}

void Cluster::Flush()
{
    for (unsigned int i=0; i<myBlockList.size(); i++)
	delete myBlockList[i];

    myBlockList.clear();

    if (myBuffer != NULL)
    {
	delete myBuffer;
	myBuffer = NULL;
	myBufferBlockPointer = NULL;
	myBufferBlockSize = 0;
    }
}

/*!
    \brief fill myBuffer with new Cluster data
    \todo handle checking the position of the cluster
    \todo use the header/footer sizes from FileKax instead of CLUSTER_HEADER_SIZE
    \todo really find the correct one (including correct checksum)
*/
uint32 Cluster::FindHead(IOCallback & input)
{
    if (myBlockList.size() != 0)
	throw 0;

    // read the head first
    uint32 offset = 0;
    int FoundIndex = 0;
    binary Header;
    uint64 FoundPosition = input.getFilePointer();
    
    while (FoundIndex != 4)
    {
	try {
	    input.readFully(&Header,1);
	    offset++;
	}
	catch (exception & Ex)
	{
	    stringstream Msg;
	    Msg << "EOF in FindHead";
	    throw runtime_error(Msg.str());
	}

	if (Header == HEADER[FoundIndex])
	    FoundIndex++;
	else {
	    FoundPosition = input.getFilePointer();
	    FoundIndex = 0;
	}
    }

    try {
	binary Head[CLUSTER_HEADER_SIZE-4];
	input.readFully(Head,CLUSTER_HEADER_SIZE-4);
	offset -= 4;

	/// \todo verify the position ? emits a warning if not correct

	// cluster size
	big_uint32 cluster_size;
	cluster_size.Eval(Head);
	mySize = cluster_size;

	// read the Block data
	myBuffer = new binary[mySize - CLUSTER_HEADER_SIZE];
	myBufferBlockPointer = myBuffer;
	myBufferBlockSize = mySize - CLUSTER_HEADER_SIZE - CLUSTER_TRAILER_SIZE;
	input.readFully(myBuffer, mySize - CLUSTER_HEADER_SIZE);

	// verify the Checksum
	Adler32 checksum;
	checksum.Feed(HEADER, 4);
	checksum.Feed(Head, CLUSTER_HEADER_SIZE-4);
	checksum.Feed(myBuffer, mySize - CLUSTER_HEADER_SIZE - CLUSTER_TRAILER_SIZE);
	if (checksum != &myBuffer[mySize - CLUSTER_HEADER_SIZE - CLUSTER_TRAILER_SIZE])
	    throw 0;
    }
    catch (exception & Ex)
    {
	throw Ex;
    }

    return offset;
}

bool Cluster::GetBlock(binary * & aBuffer, uint32 & aBufferSize, uint8 & aTrackNumber)
{
    if (myBufferBlockPointer == NULL)
	throw 0;

    aBuffer = myBufferBlockPointer;
    aBufferSize = myBufferBlockSize;
    bool result = Block::GetHeaderInfo(aBuffer, aBufferSize, aTrackNumber);
    myBufferBlockSize -= aBufferSize;
    myBufferBlockPointer += aBufferSize;

    return result;
}

END_MATROSKA_NAMESPACE
