/*
	libmatroska : parse and create Matroska files, see http://matroska.sf.net/
	Copyright (C) 2002 Steve Lhomme

	This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

/*!
	\file
	\version \$Id: ContentInfo.cpp,v 1.9 2002/05/13 08:32:32 robux4 Exp $
	\author Ingo Ralf Blum   <ingoralfblum @ users.sf.net>
	\author Lasse Krkkinen <tronic @ users.sf.net>
	\author Steve Lhomme     <robux4 @ users.sf.net>
*/
#include "StdInclude.hpp"
#include "ContentInfo.hpp"
#include "Adler32.hpp"

START_MATROSKA_NAMESPACE

ContentInfo::ContentInfo()
 :myRadioButton1(0)
 ,myRadioButton2(0)
 ,myContentSubtype(0)
 ,mySmallText(64)
 ,myTextFields1(128)
 ,myTextFields2(128)
 ,myTextFields3(128)
 ,myTextFields4(128)
 ,myTextFields5(128)
 ,myTextFields6(128)
 ,myTextFields7(128)
{
    for (int i=0; i<8; i ++)
    {
	myCheckBoxes[i] = false;
    }
}


uint32 ContentInfo::Render(IOCallback & output) const
{
    uint32 written_size = default_size();

    try {
	binary *Header = new binary[written_size];
	memset(Header, 0x00, written_size);

	// Textfields 1-7
        memcpy(&Header[0*128],myTextFields1.buffer(),myTextFields1.size());
        memcpy(&Header[1*128],myTextFields2.buffer(),myTextFields2.size());
        memcpy(&Header[2*128],myTextFields3.buffer(),myTextFields3.size());
        memcpy(&Header[3*128],myTextFields4.buffer(),myTextFields4.size());
        memcpy(&Header[4*128],myTextFields5.buffer(),myTextFields5.size());
        memcpy(&Header[5*128],myTextFields6.buffer(),myTextFields6.size());
        memcpy(&Header[6*128],myTextFields7.buffer(),myTextFields7.size());

	// Smalltext
	memcpy(&Header[7*128],mySmallText.buffer(),mySmallText.size());

	// Checkboxes 1-8
	binary BitBlock = 0;
	for (int i=0; i<8; i++)
	{
	    BitBlock |= (myCheckBoxes[i]?1:0)<<(7-i);
	}
	Header[0x3F8] = BitBlock;

	// Radio buttons
	Header[0x3F9] = myRadioButton1;
	Header[0x3FA] = myRadioButton2;

	// Content subtype
	Header[0x3FB] = myContentSubtype;

	// Checksum
	Adler32 checksum;
	checksum.Feed(&Header[0],written_size - 4);
	memcpy(&Header[written_size-4],&checksum.endian(),4);

	output.writeFully(Header,written_size);

	delete Header;
    }
    catch (exception & Ex)
    {
	throw Ex;
    }

    return written_size;
}

uint32 ContentInfo::Read(IOCallback & input)
{
    try {
	binary Header[1024];
	input.readFully(Header, default_size());
	uint32 result = default_size();

	if (result != default_size())
	    throw 0; // unexpected end of file

	// verify checksum
	Adler32 checksum;
	checksum.Feed(Header, default_size() - 4);
	if (checksum != &Header[default_size()-4])
	    throw 0;

	// Textfields 1-7
	myTextFields1 = (char*)&Header[0*128];
	myTextFields2 = (char*)&Header[1*128];
	myTextFields3 = (char*)&Header[2*128];
	myTextFields4 = (char*)&Header[3*128];
	myTextFields5 = (char*)&Header[4*128];
	myTextFields6 = (char*)&Header[5*128];
	myTextFields7 = (char*)&Header[6*128];

	// Smalltext
	mySmallText = (char*)&Header[7*128];

	// Checkboxes 1-8
	binary BitBlock = Header[0x3F8];
	for (int i=0; i<8; i++)
	{
	    myCheckBoxes[i] = (BitBlock & (1<<i)) != 0;
	}

	// Radio buttons
	myRadioButton1 = Header[0x3F9] & 0x07;
	myRadioButton2 = Header[0x3FA] & 0x07;

	// Content subtype
	myContentSubtype = Header[0x3FB];

	return result;
    }
    catch (exception & Ex)
    {
	throw Ex;
    }
}

END_MATROSKA_NAMESPACE
