/*
	libmatroska : parse and create Matroska files, see http://matroska.sf.net/
	Copyright (C) 2002 Steve Lhomme

	This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

/*!
	\file
	\version \$Id: ExtendedInfo.cpp,v 1.10 2002/05/13 08:32:32 robux4 Exp $
	\author Ingo Ralf Blum   <ingoralfblum @ users.sf.net>
	\author Lasse Krkkinen <tronic @ users.sf.net>
	\author Steve Lhomme     <robux4 @ users.sf.net>
*/
#include "StdInclude.hpp"
#include "ExtendedInfo.hpp"
#include "Adler32.hpp"

START_MATROSKA_NAMESPACE

const size_t ExtendedInfo::TITLE_SIZE = 192;
const size_t ExtendedInfo::EDITION_SIZE = 128;
const size_t ExtendedInfo::NAME_SIZE = 128;
const size_t ExtendedInfo::URL_SIZE = 128;
const size_t ExtendedInfo::EMAIL_SIZE = 64;
const size_t ExtendedInfo::ENCODER_NAME_SIZE = 128;
const size_t ExtendedInfo::ENCODER_URL_SIZE = 128;
const size_t ExtendedInfo::ENCODER_EMAIL_SIZE = 64;
const size_t ExtendedInfo::DESCRIPTION_SIZE = 2048;
const size_t ExtendedInfo::COUNTRY_CODE_SIZE = 2;
//const size_t ExtendedInfo::LANGUAGE_CODE_SIZE = 4;

ExtendedInfo::ExtendedInfo()
 :myTitle(TITLE_SIZE)
 ,myEdition(EDITION_SIZE)
 ,myName(NAME_SIZE)
 ,myURL(URL_SIZE)
 ,myEmail(EMAIL_SIZE)
 ,myEncoderName(ENCODER_NAME_SIZE)
 ,myEncoderURL(ENCODER_URL_SIZE)
 ,myEncoderEmail(ENCODER_EMAIL_SIZE)
 ,myDescription(DESCRIPTION_SIZE)
 ,myCountryCode(COUNTRY_CODE_SIZE)
// ,myLanguageCode(LANGUAGE_CODE_SIZE)
 ,myContentType(0)
 ,myEncodingDate(0)
 ,myLastEditingDate(0)
 ,myProductionYear(0)
{
    memset(myRsaci,0,RSACI_SIZE); ///< is this a valid default value for RSACI ?
    myLanguageCode[0] =  0;
    myLanguageCode[1] = 'u';
    myLanguageCode[2] = 'n';
    myLanguageCode[3] = 'd';
}

/*!
    \todo handle data properly
    \todo the myLastEditingDate could be computed on the fly
*/
uint32 ExtendedInfo::Render(IOCallback & output) const
{
    uint32 written_size = default_size();

    try {
	binary *Header = new binary[written_size];
	memset(Header, 0x00, written_size);

	binary *cursor = Header;
    
	memcpy(cursor,myTitle.buffer(),TITLE_SIZE);
	cursor += TITLE_SIZE;

	memcpy(cursor,myEdition.buffer(),EDITION_SIZE);
	cursor += EDITION_SIZE;

	memcpy(cursor,myName.buffer(),NAME_SIZE);
	cursor += NAME_SIZE;

	memcpy(cursor,myURL.buffer(),URL_SIZE);
	cursor += URL_SIZE;

	memcpy(cursor,myEmail.buffer(),EMAIL_SIZE);
	cursor += EMAIL_SIZE;

	memcpy(cursor,myEncoderName.buffer(),ENCODER_NAME_SIZE);
	cursor += ENCODER_NAME_SIZE;

	memcpy(cursor,myEncoderURL.buffer(),ENCODER_URL_SIZE);
	cursor += ENCODER_URL_SIZE;

	memcpy(cursor,myEncoderEmail.buffer(),ENCODER_EMAIL_SIZE);
	cursor += ENCODER_EMAIL_SIZE;

	memcpy(cursor,myDescription.buffer(),DESCRIPTION_SIZE);
	cursor += DESCRIPTION_SIZE;

	// RSACI rating
	memcpy(cursor,myRsaci,RSACI_SIZE);
	cursor += RSACI_SIZE;

	// Flags (nothing yet)
	*cursor++ = 0;

	// Content Type 
	*cursor++ = myContentType; ///< \see http://matroska.sourceforge.net/specs/

	// File access dates
	big_uint32 date = myEncodingDate;
	date.Fill(cursor);
	cursor += sizeof(uint32);

	date = myLastEditingDate;
	date.Fill(cursor);
	cursor += sizeof(uint32);

	// Production date
	big_uint16 production = myProductionYear;
	production.Fill(cursor);
	cursor += sizeof(16);

	// Language code
	memcpy(cursor, &myLanguageCode, countof(myLanguageCode));
	cursor += countof(myLanguageCode);

	// Checksum
	Adler32 checksum;
	checksum.Feed(&Header[0],written_size - 4);
	memcpy(&Header[written_size-4],&checksum.endian(),4);

	output.writeFully(Header,written_size);
    }
    catch (exception & Ex)
    {
	throw Ex;
    }

    return written_size;
}

uint32 ExtendedInfo::Read(IOCallback & input)
{
    try {
	binary Header[3072];
	input.readFully(Header, default_size());
	uint32 result = default_size();

	if (result != default_size())
	    throw 0; // unexpected end of file

	// verify checksum
	Adler32 checksum;
	checksum.Feed(Header, default_size() - 4);
	if (checksum != &Header[default_size()-4])
	    throw 0;

	char *cursor = (char*) Header;

	myTitle = cursor;
	cursor += TITLE_SIZE;

	myEdition = cursor;
	cursor += EDITION_SIZE;

	myName = cursor;
	cursor += NAME_SIZE;

	myURL = cursor;
	cursor += URL_SIZE;

	myEmail = cursor;
	cursor += EMAIL_SIZE;

	myEncoderName = cursor;
	cursor += ENCODER_NAME_SIZE;

	myEncoderURL = cursor;
	cursor += ENCODER_URL_SIZE;

	myEncoderEmail = cursor;
	cursor += ENCODER_EMAIL_SIZE;

	myDescription = cursor;
	cursor += DESCRIPTION_SIZE;

	// RSACI rating
	memcpy(myRsaci,cursor,RSACI_SIZE);
	cursor += RSACI_SIZE;

	// flags
	cursor++;

	myContentType = *cursor++;

	// File access dates
	big_uint32 date;
	date.Eval((binary*) cursor);
	myEncodingDate = date;
	cursor += sizeof(uint32);

	date.Eval((binary*) cursor);
	myLastEditingDate = date;
	cursor += sizeof(uint32);

	big_uint16 year;
	year.Eval((binary*) cursor);
	myProductionYear = year;
	cursor += sizeof(uint16);

	// country code
	myCountryCode = cursor;
	cursor += COUNTRY_CODE_SIZE;

	// language code
	for (int i=0; i<4; i ++)
	    myLanguageCode[i] = *cursor++;

	return result;
    }
    catch (exception & Ex)
    {
	throw Ex;
    }
}

END_MATROSKA_NAMESPACE
