/*
	libmatroska : parse and create Matroska files, see http://matroska.sf.net/
	Copyright (C) 2002 Ingo Ralf Blum

	This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

/*!
	\file
	\version \$Id: StdIOCallback.cpp,v 1.7 2002/05/17 18:02:18 robux4 Exp $
*/
#include "StdInclude.hpp"
#include "StdIOCallback.hpp"
#include "Debug.hpp"
#include "Config.hpp"

#include <cassert>
#include <sstream>
#include <climits>

using namespace std;

START_MATROSKA_NAMESPACE

CRTError::CRTError(int Error,const std::string&Description)
	:std::runtime_error(Description+": "+strerror(Error))
	,Error(Error)
{
}

CRTError::CRTError(const std::string&Description,int Error)
	:std::runtime_error(Description+": "+strerror(Error))
	,Error(Error)
{
}

StdIOCallback::StdIOCallback(const char*Path, const open_mode aMode)
{
    assert(Path!=0);

    char *Mode;
    switch (aMode)
    {
	case MODE_READ:
	    Mode = "rb";
	    break;
	case MODE_WRITE:
	    Mode = "wb";
	    break;
	case MODE_CREATE:
	    Mode = "wb+";
	    break;
	default:
	    throw 0;
    }

    File=fopen(Path,Mode);
    if(File==0)
    {
	stringstream Msg;
	Msg<<"Can't open stdio file \""<<Path<<"\" in mode \""<<Mode<<"\"";
	throw CRTError(Msg.str());
    }

    MATROSKA_TRACE("Successfully opened file \"%hs\" in mode \"%hs\". The handle is %p\n",Path,Mode,File);
}

StdIOCallback::~StdIOCallback()throw()
{
	try
	{
		close();
	}
	catch(std::exception&Ex)
	{
		MATROSKA_TRACE("Error in ~StdIOCallback: %hs\n",Ex.what());
	}
	catch(...)
	{
		MATROSKA_TRACE("Unknown error in ~StdIOCallback\n");
	}
}

uint32 StdIOCallback::read(void*Buffer,size_t Size)
{
    assert(File!=0);

    size_t result = fread(Buffer, 1, Size, File);
    return result;
}

void StdIOCallback::setFilePointer(int64 Offset,seek_mode Mode)
{
	assert(File!=0);

	// There is a numeric cast in the boost library, which would be quite nice for this checking
/* 
	SL : replaced because unknown class in cygwin

	assert(Offset<=numeric_limits<long>::max());
	assert(Offset>=numeric_limits<long>::min());
*/
	assert(Offset <= LONG_MAX);
	assert(Offset >= LONG_MIN);

	assert(Mode==SEEK_CUR||Mode==SEEK_END||Mode==SEEK_SET);

	if(fseek(File,Offset,Mode)!=0)
	{
		ostringstream Msg;
		Msg<<"Failed to seek file "<<File<<" to offset "<<(unsigned long)Offset<<" in mode "<<Mode;
		throw CRTError(Msg.str());
	}
}

size_t StdIOCallback::write(const void*Buffer,size_t Size)
{
	assert(File!=0);

	return fwrite(Buffer,1,Size,File);
}

uint64 StdIOCallback::getFilePointer()
{
	assert(File!=0);

	long Result=ftell(File);
	if(Result<0)
	{
		stringstream Msg;
		Msg<<"Can't tell the current file pointer position for "<<File;
		throw CRTError(Msg.str());
	}

	return Result;
}

void StdIOCallback::close()
{
	if(File==0)
		return;

	if(fclose(File)!=0)
	{
		stringstream Msg;
		Msg<<"Can't close file "<<File;
		throw CRTError(Msg.str());
	}

	File=0;
}

END_MATROSKA_NAMESPACE
