/*
	libmatroska : parse and create Matroska files, see http://matroska.sf.net/
	Copyright (C) 2002 Steve Lhomme

	This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

/*!
	\file
	\version \$Id: Track.cpp,v 1.17 2002/11/20 22:37:39 robux4 Exp $
	\author Ingo Ralf Blum   <ingoralfblum @ users.sf.net>
	\author Lasse Krkkinen <tronic @ users.sf.net>
	\author Steve Lhomme     <robux4 @ users.sf.net>
*/
#include "StdInclude.hpp"
#include "Track.hpp"
#include "Adler32.hpp"
#include "BlockFxS.hpp"
#include "BlockOcL.hpp"
#include "BlockSPB.hpp"
#include "TrackAudio.hpp"
#include "TrackVideo.hpp"

#include <sstream>

START_MATROSKA_NAMESPACE

using namespace std;

const character Track::HEADER[4] = { 'T', 'r', 'k', 'E' };

Track::Track(const track_type aType)
 :myHasBlocks(false)
 ,myBlock(NULL)
 ,myTrackNumber(0)
 ,myNumberLocked(false)
 ,myEnabled(true)
 ,myPreferred(true)
 ,myBlockSize(0)
 ,myLaced(false)
 ,myFormat(TRACK_SIZE_FORMAT)
 ,myVersion1(0)
 ,myVersion2(0)
 ,myVersion3(0)
 ,myCodecHeader(NULL)
 ,myCodecHeaderSize(0)
 ,myCodecName(TRACK_SIZE_CODEC_NAME)
 ,myCodecUrl(TRACK_SIZE_CODEC_URL1)
 ,myCodecUrl_alt(TRACK_SIZE_CODEC_URL2)
 ,myNanoSecondPerBlock(0)
 ,myCodecSettings(TRACK_SIZE_SETTINGS)
 ,myName(TRACK_SIZE_NAME)
{
    myTrackType = TrackType::GetNewTrack(aType);

    if (myTrackType == NULL)
	throw 0; // we can't work with a NULL tracktype

    myLanguageCode[0] =  0;
    myLanguageCode[1] = 'u';
    myLanguageCode[2] = 'n';
    myLanguageCode[3] = 'd';
}

Track::~Track()
{
    if (myTrackType != NULL)
	delete myTrackType;
}

uint32 Track::RenderEntry(IOCallback & output, const uint8 aTrackNumber)
{
    uint32 written_size = default_size();

    try {
	if (myNumberLocked)
	    throw 0; // you have to manually unlock it

	if (myTrackType == NULL)
	    throw 0;

	if (myCodecHeaderSize != 0)
	    throw 0; ///< Codec Header writing not supported yet

	binary *Header = new binary[written_size];
	memset(Header, 0x00, written_size);
	binary *cursor = Header;

	*cursor++ = HEADER[0];
	*cursor++ = HEADER[1];
	*cursor++ = HEADER[2];
	*cursor++ = HEADER[3];

	// track type
	*cursor++ = track_type(*myTrackType) & 0xFF;

	// flags
	*cursor++ = (myEnabled ? 0x80:0) | (myPreferred ? 0x40:0) | (myBlockSize!=0 ? 0x20:0) | (myLaced ? 0x10:0);

	// seek entries
	cursor = &Header[0x08];
	big_uint32 bu32 = 0; // not handled for the moment
	bu32.Fill(cursor);
	cursor += sizeof(uint32);

	// language code
	*cursor++ = myLanguageCode[0];
	*cursor++ = myLanguageCode[1];
	*cursor++ = myLanguageCode[2];
	*cursor++ = myLanguageCode[3];

	// format string
	memcpy(cursor, myFormat.buffer(), myFormat.size());
	cursor += myFormat.size();

	// versions
	big_uint16 _version;
	_version = myVersion1;
	_version.Fill(cursor);
	cursor += 2;

	_version = myVersion2;
	_version.Fill(cursor);
	cursor += 2;

	_version = myVersion3;
	_version.Fill(cursor);
	cursor += 2;

	// size of codec header
	cursor = &Header[0x28];
	bu32 = myCodecHeaderSize; // not handled for the moment
	bu32.Fill(cursor);
	cursor += sizeof(uint32);

	// codec header checksum
	Adler32 _checksum;
	_checksum.Feed(myCodecHeader,myCodecHeaderSize);
	memcpy(cursor,&_checksum.endian(),4);
	cursor += 4;

	// codec name
	memcpy(cursor, myCodecName.buffer(), myCodecName.size());
	cursor += myCodecName.size();

	// codec URL
	memcpy(cursor, myCodecUrl.buffer(), myCodecUrl.size());
	cursor += myCodecUrl.size();

	// codec alternative URL
	memcpy(cursor, myCodecUrl_alt.buffer(), myCodecUrl_alt.size());
	cursor += myCodecUrl_alt.size();

	// nano seconds per block
	big_uint64 _NanoSecond = myNanoSecondPerBlock;
	_NanoSecond.Fill(cursor);
	cursor += 8;

	// encoding settings
	cursor = &Header[0x100];
	memcpy(cursor, myCodecSettings.buffer(), myCodecSettings.size());
	cursor += myCodecSettings.size();

	// track name
	memcpy(cursor, myName.buffer(), myName.size());
	cursor += myName.size();

	// track-type dependent
	myTrackType->Render(cursor);

	// Checksum
	Adler32 checksum;
	checksum.Feed(&Header[0],written_size - 4);
	memcpy(&Header[written_size-4],&checksum.endian(),4);

	output.writeFully(Header,written_size);

	myTrackNumber = aTrackNumber;
	myNumberLocked = true;
    }
    catch (exception & Ex)
    {
	throw Ex;
    }

    return written_size;
}

/*!
    \return wether we should render the current block or not
*/
bool Track::AddFrame(const uint32 aTimecode,
		     const binary *aFrame, const uint32 aFrameSize, const bool KeyFrame,
		     const bool BFrame)
{
    bool result = true;
    try {
        // create the Block
	if (myBlock == NULL) {
	    if (myLaced)
		myBlock = new BlockOcL(myTrackNumber, aTimecode);
	    else if (myBlockSize != 0)
		myBlock = new BlockFxS(myTrackNumber, aTimecode, myBlockSize);
	    else
		myBlock = new BlockSPB(myTrackNumber, aTimecode);

	    if (myBlock == NULL) {
		stringstream Msg;
		Msg << "No more memory left";
		throw runtime_error(Msg.str()); // we only allow frames of the same length
	    }
	}
    
	return myBlock->AddFrame(aFrame, aFrameSize, KeyFrame, BFrame);
    }
    catch (exception & Ex)
    {
	throw Ex;
    }

    return result; // wether we can add more data to the Cluster
}

bool Track::SerialiseBlock(Cluster & aCluster)
{
    if (myBlock == NULL)
	throw 0;

    if (aCluster.AddBlock(myBlock))
    {
	myBlock = NULL; // the track is not the owner of this block anymore
	return true;
    }

    return false;
}

void Track::SetName(const std::string & aStr) 
{
    if (myHasBlocks)
	throw 0;
    myName = aStr;
}

void Track::SetLaced(const bool bLaced)
{
    if (myHasBlocks)
	throw 0;
    myLaced = bLaced;
}

/**
    \todo handle myBlockSize reading
*/
Track * Track::ReadEntry(IOCallback & input, const uint8 aTrackNumber, const StreamInfo & aStreamInfo)
{
    if (aStreamInfo.TrackEntrySize < default_size())
	throw 0;

    try {
	binary *Buffer = new binary[aStreamInfo.TrackEntrySize];

	input.readFully(Buffer, aStreamInfo.TrackEntrySize);

	// verify checksum
	Adler32 checksum;
	checksum.Feed(Buffer, aStreamInfo.TrackEntrySize - 4);
	if (checksum != &Buffer[aStreamInfo.TrackEntrySize - 4])
	    throw 0;

	binary *cursor = Buffer;

	// check wether it's a valid track type
	for (int i = 0; i < 4; i ++)
	{
    	    if (*cursor++ != HEADER[i])
		throw 0; // this is not a Track Entry !!!
	}

	// check the Track Type and create the appropriate class
	track_type type = track_type(*cursor++);
	
	Track *result = new Track(type);

	if (result == NULL)
	    throw 0;
    
	result->myNumberLocked = true;
	result->myTrackNumber = aTrackNumber;

	// flags
	result->myEnabled = (*cursor & 0x80) != 0;
	result->myPreferred = (*cursor & 0x40) != 0;
	result->myLaced = (*cursor & 0x10) != 0;

	cursor = &Buffer[0x0C];
	// language code
	for (i=0; i<4; i ++)
	    result->myLanguageCode[i] = *cursor++;

	// format string
	result->myFormat = (char*) cursor;
	cursor += result->myFormat.size();

	// version numbers
	big_uint16 bu16;
	bu16.Eval(cursor);
	cursor += sizeof(uint16);
	result->myVersion1 = bu16;

	bu16.Eval(cursor);
	cursor += sizeof(uint16);
	result->myVersion2 = bu16;

	bu16.Eval(cursor);
	cursor += sizeof(uint16);
	result->myVersion3 = bu16;

	cursor = &Buffer[0x28];
	// Codec header size
	big_uint32 bu32;
	bu32.Eval(cursor);
	cursor += sizeof(uint32);
	result->myCodecHeaderSize = bu32;

	/// \todo handle the codec header Checksum

	cursor = &Buffer[0x30];
	// Codec information
	result->myCodecName = (char*) cursor;
	cursor += result->myCodecName.size();

	result->myCodecUrl = (char*) cursor;
	cursor += result->myCodecUrl.size();

	result->myCodecUrl_alt = (char*) cursor;
	cursor += result->myCodecUrl_alt.size();

	// nanoseconds per block
	big_uint64 bu64;
	bu64.Eval(cursor);
	cursor += sizeof(uint64);
	result->myNanoSecondPerBlock = bu64;

	cursor = &Buffer[0x100];
	// encoding settings
	result->myCodecSettings = (char*) cursor;
	cursor += result->myCodecSettings.size();

	// Track name
	result->myName = (char*) cursor;
	cursor += result->myName.size();

	result->myTrackType->Read(cursor);

	return result;
    }
    catch (exception & Ex)
    {
	throw Ex;
    }
}

void Track::SetInfoFormat(const std::string aFormat)
{
	myFormat = aFormat;
}

void Track::GetInfo(TrackInfo & aTrackInfo) const
{
    aTrackInfo.TrackType = track_type(*myTrackType);
    aTrackInfo.Enabled = myEnabled;
    aTrackInfo.Preferred = myPreferred;
//    aTrackInfo.BlockSize = my; ///< 0 if block size is not fixed

    for (int i=0;i<4;i++)
	aTrackInfo.LanguageCode[i] = myLanguageCode[i];

    aTrackInfo.Format = (char*) myFormat.buffer();
    aTrackInfo.Version1 = myVersion1; ///< encoding
    aTrackInfo.Version2 = myVersion2; ///< reading
    aTrackInfo.Version3 = myVersion3; ///< absolute minimum reading

    aTrackInfo.CodecName = (char*) myCodecName.buffer();
    aTrackInfo.CodecUrl = (char*) myCodecUrl.buffer();
    aTrackInfo.CodecUrl_alt = (char*) myCodecUrl_alt.buffer();

    aTrackInfo.NanoSecondPerBlock = myNanoSecondPerBlock;

    aTrackInfo.CodecSettings = (char*) myCodecSettings.buffer();

    aTrackInfo.Name = (char*) myName.buffer();
}

// Audio related getters/setters

void Track::GetInfoAudio(TrackInfoAudio & aTrackInfo) const
{
    if (track_type(*this) != track_audio)
	throw 0;

    aTrackInfo = *reinterpret_cast<TrackAudio*>(myTrackType);
}

void Track::SetInfoAudio(const TrackInfoAudio & aTrackInfo)
{
    if (track_type(*this) != track_audio)
	throw 0;

    reinterpret_cast<TrackAudio*>(myTrackType)->SetInfo(aTrackInfo);
}

// Video related getters/setters

void Track::GetInfoVideo(TrackInfoVideo & aTrackInfo) const
{
    if (track_type(*this) != track_video)
	throw 0;

    aTrackInfo = *reinterpret_cast<TrackVideo*>(myTrackType);
}

void Track::SetInfoVideo(const TrackInfoVideo & aTrackInfo)
{
    if (track_type(*this) != track_video)
	throw 0;

    reinterpret_cast<TrackVideo*>(myTrackType)->SetInfo(aTrackInfo);
}

void Track::HandleBlock(const binary * aFrame, const uint32 aFrameSize)
{
    if (myBlock != NULL)
	throw 0; // we can't work like this

    if (myLaced)
	myBlock = new BlockOcL(myTrackNumber, aFrame, aFrameSize);
    else if (myBlockSize != 0)
	myBlock = new BlockFxS(myTrackNumber, myBlockSize, aFrame, aFrameSize);
    else
	myBlock = new BlockSPB(myTrackNumber, aFrame, aFrameSize);

    if (myBlock == NULL) {
	stringstream Msg;
	Msg << "No more memory left";
	throw runtime_error(Msg.str()); // we only allow frames of the same length
    }
}

void Track::GetNextFrame(uint32 & aTimecode, const Frame * & aFrame,
						 bool & aKeyFrame, bool & aBFrame)
{
    if (myBlock == NULL)
	throw 0;

    myBlock->GetNextFrame(aTimecode, aFrame, aKeyFrame, aBFrame);
}

bool Track::NoFrameLeft() const
{
    if (myBlock == NULL)
	throw 0;

    return myBlock->NoFrameLeft();
}

void Track::FlushBlock()
{
    if (myBlock != NULL)
    {
	delete myBlock;
	myBlock = NULL;
    }
}

/*!
    \todo Support codec header writing (currently size = 0)
*/
uint32 Track::RenderCodecHeader(IOCallback & output) const
{
    if (!myNumberLocked)
	throw 0;

    if (myCodecHeaderSize != 0)
	throw 0; ///< \todo Not supported for the moment

    return myCodecHeaderSize;
}

uint32 Track::ReadCodecHeader(IOCallback & input)
{
    if (myCodecHeaderSize == 0)
	return 0;

    if (myCodecHeader != NULL)
	delete myCodecHeader;

    myCodecHeader = new binary[myCodecHeaderSize];
    input.readFully(myCodecHeader, myCodecHeaderSize);

    return myCodecHeaderSize;
}

END_MATROSKA_NAMESPACE
