/*
	libmatroska : parse and create Matroska files, see http://matroska.sourceforge.net/
	Copyright (C) 2002 Steve Lhomme

	This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

/*!
	\file
	\version \$Id: Track.hpp,v 1.15 2002/11/20 22:37:39 robux4 Exp $
	\author Ingo Ralf Blum   <ingoralfblum @ users.sf.net>
	\author Lasse Krkkinen <tronic @ users.sf.net>
	\author Steve Lhomme     <robux4 @ users.sf.net>
*/
#ifndef LIBMATROSKA_TRACK_HPP
#define LIBMATROSKA_TRACK_HPP

#include "Config.hpp"
#include "Types.hpp"
#include "TruncString.hpp"
#include "IOCallback.hpp"
#include "Cluster.hpp"
#include "StreamInfo.hpp"
#include "TrackType.hpp"

START_MATROSKA_NAMESPACE

static const size_t TRACK_SIZE_FORMAT = 16;
static const size_t TRACK_SIZE_CODEC_NAME = 64;
static const size_t TRACK_SIZE_CODEC_URL1 = 64;
static const size_t TRACK_SIZE_CODEC_URL2 = 64;
static const size_t TRACK_SIZE_SETTINGS = 64;
static const size_t TRACK_SIZE_NAME = 64;

class Block;
class Frame;

/*!
    \class Track
    \brief Describe the data of a track (referenced by a track number)
    \see http://matroska.sourceforge.net/specs/#track
*/
class Track {
    public:
	Track(const track_type aType);
	Track(const Track & copy);
	~Track();

	uint8 TrackNumber() const { return myTrackNumber; }
	operator track_type() const
	{ 
	    if (myTrackType == NULL)
		return track_type(track_unknown);
	    else return *myTrackType;
	}

	void SetName(const std::string & aStr);
	void SetLaced(const bool bLaced = true);
	static inline uint32 default_size() {return 576;}

	uint32 RenderEntry(IOCallback & output, const uint8 aTrackNumber);
	static Track * ReadEntry(IOCallback & input, const uint8 aTrackNumber, const StreamInfo & aStreamInfo);

	uint32 RenderCodecHeader(IOCallback & output) const;
	uint32 ReadCodecHeader(IOCallback & input);

	bool AddFrame(const uint32 timecode,
		      const binary *aFrame, const uint32 aFrameSize,
		      const bool KeyFrame = true, const bool BFrame = false);

	bool ReadFrame(IOCallback & input, const Cluster & a_cluster,
		      binary *aFrame, uint32 & aFrameSize,
		      bool & KeyFrame, bool & BFrame);

	void SetInfoFormat(const std::string aFormat);
	void GetInfo(TrackInfo & aTrackInfo) const;
	
	void GetInfoAudio(TrackInfoAudio & aTrackInfo) const;
	void SetInfoAudio(const TrackInfoAudio & aTrackInfo);

	void GetInfoVideo(TrackInfoVideo & aTrackInfo) const;
	void SetInfoVideo(const TrackInfoVideo & aTrackInfo);

	void HandleBlock(const binary * aFrame, const uint32 aFrameSize);
	void GetNextFrame(uint32 & aTimecode, const Frame * & aFrame,
		      bool & aKeyFrame, bool & aBFrame);
	bool NoFrameLeft() const;
	void FlushBlock();
	
	bool SerialiseBlock(Cluster & aCluster);

	uint32 CodecSize() const {return myCodecHeaderSize;}

    protected:
	bool myHasBlocks;

        uint8 myTrackNumber;
	bool myNumberLocked;
	TrackType * myTrackType; // generated on the fly

	static const character HEADER[4];

	bool myEnabled;
	bool myPreferred;
	uint32 myBlockSize; ///< 0 if block size is not fixed
	bool myLaced;

	binary myLanguageCode[4]; ///< \see http://matroska.sourceforge.net/language.htm

	TruncString myFormat;
	uint16 myVersion1; ///< encoding
	uint16 myVersion2; ///< reading
	uint16 myVersion3; ///< absolute minimum reading

	binary     *myCodecHeader;
	uint32      myCodecHeaderSize;

	TruncString myCodecName;
	TruncString myCodecUrl;
	TruncString myCodecUrl_alt;

	uint64 myNanoSecondPerBlock;

	TruncString myCodecSettings;

	TruncString8 myName;

	Block * myBlock;
};

END_MATROSKA_NAMESPACE

#endif // LIBMATROSKA_TRACK_HPP
