/*
	libmatroska : parse and create Matroska files, see http://matroska.sourceforge.net/
	Copyright (C) 2002 Steve Lhomme

	This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

/*!
    \file
    \version \$Id: TruncString.hpp,v 1.5 2002/05/08 16:31:17 robux4 Exp $
    \author Ingo Ralf Blum   <ingoralfblum @ users.sf.net>
    \author Lasse Krkkinen <tronic @ users.sf.net>
    \author Steve Lhomme     <robux4 @ users.sf.net>
*/
#ifndef LIBMATROSKA_TRUNC_STRING_HPP
#define LIBMATROSKA_TRUNC_STRING_HPP

#include <string>

#include "Config.hpp"
#include "Types.hpp"

START_MATROSKA_NAMESPACE

/*!
	\class TruncString
	\brief Handle strings that has a fixed size when rendered
*/
class TruncString {
    public:
	enum align {
	    left,
	    right
	};

	/*!
	    \brief Create a string truncated to the output size
	    \param limit   The size for which to truncate the string
	    \param str     An ASCII/latin-1 string use to initialise the buffer
	    \param a_align Possibility to align the string on the left or the right of the buffer
	    \param a_pad   The value used to fill the unused parts in the string (default to 0x00)
	*/
	TruncString(size_t limit, const std::string & str = "", const align a_align = left, const char a_pad = '\0');
	~TruncString();

	size_t size() const {return myLimit;}

	/*!
	    \brief Fill the buffer with a truncated string
	    \param str     An ASCII/latin-1 string use to fill the buffer
	*/
	TruncString &operator=(const std::string & aStr);

	const binary *buffer() const {return myBuffer;}

    protected:
	size_t myLimit;
	std::string myString;
	binary *myBuffer;
	align myAlignement;
	const char myPad;

	void Fill(const std::string & aStr);
};

/*!
	\class TruncString8
	\brief Handle UTF-8 strings that has a fixed size when rendered
	\todo Handle feeding the class with wide strings
	\note The data are stored natively in UTF8, when fed with UTF16 it's converted on the fly
*/
class TruncString8 {
    public:
	enum align {
	    left,
	    right
	};

	/*!
	    \brief Create a string truncated to the output size
	    \param limit   The size for which to truncate the string
	    \param str     A UTF-8 formated string use to initialise the buffer
	    \param a_align Possibility to align the string on the left or the right of the buffer
	    \param a_pad   The value used to fill the unused parts in the string (default to 0x00)
	*/
	TruncString8(size_t limit, const std::string & str = "", const align a_align = left, const char a_pad = '\0');
//	TruncString8(const std::wstring & str,size_t limit, const align a_align = left);
	~TruncString8();

	size_t size() const {return myLimit;}

	/*!
	    \brief Fill the buffer with a truncated string
	    \param str An UTF-8 formated string use to fill the buffer
	*/
	TruncString8 &operator=(const std::string & aStr);
//	TruncString8 &operator=(const std::wstring & aStr);

	const binary *buffer() const {return myBuffer;}

    protected:
	size_t myLimit;
	std::string myString;
	binary *myBuffer;
	align myAlignement;
	const char myPad;

//	void Fill(const std::wstring & aStr);
	void Fill(const std::string & aStr);
};

END_MATROSKA_NAMESPACE

#endif // LIBMATROSKA_INTERLEAVED_DATA_HPP
