/*
	libmatroska : parse and create Matroska files, see http://matroska.sf.net/
	Copyright (C) 2002 Ingo Ralf Blum

	This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

/*!
	\file
	\version \$Id: Types.cpp,v 1.4 2002/04/14 10:03:01 robux4 Exp $
*/
#include "StdInclude.hpp"
#include "Types.hpp"

#include <cassert>

using namespace std;

START_MATROSKA_NAMESPACE

// Contains the high part of the character
static const utf16 surrogate_high_beginning=0xd800; // Including
static const utf16 surrogate_high_end=0xdc00; // Excluding

// Contains the low part of the character
static const utf16 surrogate_low_beginning=0xdc00; // Including
static const utf16 surrogate_low_end=0xe000; // Excluding

static const size_t half_shift=10;
static const utf32 half_base=0x10000;
static const utf32 half_mask=0x3ff;

static const utf32 utf16_max=0x10ffff;

// Conversion helpers
/* incompatible with cygwin
codecvt_base::result fromUTF(const utf16*SourceBeginning,const utf16*SourceEnd,const utf16*&SourceNext,utf32&Destination)
{
	assert(SourceBeginning!=0);
	assert(SourceEnd!=0);

	SourceNext=SourceBeginning;

	if(SourceBeginning>SourceEnd)
		return codecvt_base::ok; // Nothing to do

	utf16 Char=*SourceBeginning++;
	if(Char>=surrogate_high_beginning&&Char<surrogate_high_end)
	{
		if(SourceBeginning>=SourceEnd)
			return codecvt_base::partial;

		// The next character must be a low surrogate
		utf16 Low=*SourceBeginning++;
		if(Low<surrogate_low_beginning||Low>=surrogate_low_end)
			return codecvt_base::error; // A high surrogate must be followed by a low surrogate

		Destination=((Char-surrogate_high_beginning)<<half_shift)
			|((Low-surrogate_low_beginning)+half_base);
		SourceNext=SourceBeginning;
		return codecvt_base::ok;
	}
	else if(Char>=surrogate_low_beginning&&Char<surrogate_low_end)
	{
		// High part is missing
		return codecvt_base::error;
	}
	else
	{
		Destination=Char;
		SourceNext=SourceBeginning;
		return codecvt_base::ok;
	}
}

codecvt_base::result MATROSKA_EXPORT toUTF(utf32 Source,utf16*DestinationBeginning,utf16*DestinationEnd,utf16*&DestinationNext)
{
	assert(DestinationBeginning!=0);
	assert(DestinationEnd!=0);
	assert(DestinationNext!=0);

	DestinationNext=DestinationBeginning;

	if(DestinationEnd<=DestinationBeginning)
		return codecvt_base::partial;

	if(Source>utf16_max)
		return codecvt_base::error;

	// If the character is in the surrogate range, we have to split it into two characters
	if(Source>=surrogate_high_beginning&&Source<surrogate_low_end)
	{
		if((DestinationEnd-DestinationBeginning)<2)
			return codecvt_base::partial;

		*DestinationNext++=(Source>>half_shift)+surrogate_high_beginning;
		*DestinationNext++=(Source&half_mask)+surrogate_low_beginning;
	}
	else
		*DestinationNext++=Source;

	return codecvt_base::ok;
}

codecvt_base::result fromUTF(const utf8*SourceBeginning,const utf8*SourceEnd,const utf8*&SourceNext,utf32&Destination)
{
	assert(SourceBeginning!=0);
	assert(SourceEnd!=0);

	if(SourceBeginning>=SourceEnd)
		return codecvt_base::ok; // Nothing to do

	utf8 Char=*SourceBeginning++;
	if((Char&0x80)==0)
	{
		Destination=Char&0xff;
		SourceNext=SourceBeginning;
		return codecvt_base::ok;
	}
	else
	{
		// TODO: We can make a speedup here, when using a 256 entry large table, which saves us the bit masking.
		// Additionally, instead of shifting, we could use a table with offset values and use the addition operation instead
		// of the continuous shifting.
		struct
		{
			utf8 Mask;
			utf8 Marker;
			uint8 ByteCount;
		}Info[]=
		{
			// The casts are required because utf8 is a signed char
			{(utf8)0xe0,(utf8)0xc0,1}
			,{(utf8)0xf0,(utf8)0xe0,2}
			,{(utf8)0xf8,(utf8)0xf0,3}
			,{(utf8)0xfc,(utf8)0xf0,4}
			,{(utf8)0xfe,(utf8)0xfc,5}
		};

		for(uint8 i=0;i<countof(Info);++i)
		{
			if((Char&Info[i].Mask)==Info[i].Marker)
			{
				// Now try converting. We don't check the size here, but in the next loop,
				// because this way we can check the validity in the source, even if it is
				// not complete.
				Destination=Char&~Info[i].Mask; // Start with the high bits
				for(uint8 c=0;c<Info[i].ByteCount;++c)
				{
					if(SourceBeginning>=SourceEnd)
						return codecvt_base::partial;

					Char=*SourceBeginning++;
					if((Char&0xc0)!=0x80)
						return codecvt_base::error; // The byte has not the 10xxxxxx signature.

					Destination=(Destination<<6)|(Char&0x3f);
				}

				SourceNext=SourceBeginning;
				return codecvt_base::ok;
			}
		}

		// We found a bad UTF8 marker
		return codecvt_base::error;
	}
}

codecvt_base::result toUTF(utf32 Source,utf8*DestinationBeginning,utf8*DestinationEnd,utf8*&DestinationNext)
{
	DestinationNext=DestinationBeginning;

	if(DestinationEnd<=DestinationBeginning)
		return codecvt_base::partial;

	if(Source<0x80)
	{
		// A one byte code
		*DestinationNext++=Source;
		return codecvt_base::ok;
	}

	// Source is a multi byte code
	static const struct
	{
		utf32 InitialMask,FollowingMask,Max;
		uint8 ByteCount; // Number of following bytes
		utf8 Marker;
	}
	Info[]=
	{
		{0x7c0,0x3f,0x800,1,(utf8)0xc0}
		,{0xf000,0xfc0,0x10000,2,(utf8)0xe0}
		,{0x1c0000,0x3f000,0x200000,3,(utf8)0xf0}
		,{0x3000000,0xfc0000,0x4000000,4,(utf8)0xf8}
		,{0x40000000,0x3f000000,0x80000000,5,(utf8)0xfc}
	};

	for(uint8 i=0;i<countof(Info);++i)
	{
		if(Source<Info[i].Max)
		{
			if((DestinationEnd-DestinationBeginning)<(Info[i].ByteCount+1))
				return codecvt_base::partial;

			*DestinationNext++=((Source&Info[i].InitialMask)>>(6*Info[i].ByteCount))|Info[i].Marker; // Write the initial byte
			for(uint8 c=0;c<Info[i].ByteCount;++c)
				*DestinationNext++=((Source&(Info[i].FollowingMask>>(6*c)))>>(6*(Info[i].ByteCount-c-1)))|0x80; // Write the following bytes

			return codecvt_base::ok;
		}
	}

	assert((Source&0x80000000)!=0);
	return codecvt_base::error; // We only support up to 31 bits
}

UTF8CodeConverter::result UTF8CodeConverter::do_in(
	UTF8CodeConverter::state_type&State
	,const UTF8CodeConverter::to_type*SourceBeginning
	,const UTF8CodeConverter::to_type*SourceEnd
	,const UTF8CodeConverter::to_type*&SourceNext
	,UTF8CodeConverter::from_type*DestBeginning
	,UTF8CodeConverter::from_type*DestEnd
	,UTF8CodeConverter::from_type*&DestNext
	)const
{
	SourceNext=SourceBeginning;
	DestNext=DestBeginning;

	result Result=ok;
	utf32 Char;
	to_type*Next;
	while(SourceNext<SourceEnd&&
		(Result=fromUTF(SourceNext,SourceEnd,Next,Char))==ok
		)
	{
		Result=toUTF(Char,DestNext,DestEnd,DestNext);
		if(Result!=ok)
			return Result;

		// Conversion was ok. The DestNext field is already updated, so the only thing
		// left is the source buffer pointer.
		SourceNext=Next;
	}

	return Result;
}

UTF8CodeConverter::result UTF8CodeConverter::do_out(
	UTF8CodeConverter::state_type&State
	,const UTF8CodeConverter::from_type*SourceBeginning
	,const UTF8CodeConverter::from_type*SourceEnd
	,const UTF8CodeConverter::from_type*&SourceNext
	,UTF8CodeConverter::to_type*DestBeginning
	,UTF8CodeConverter::to_type*DestEnd
	,UTF8CodeConverter::to_type*&DestNext
	)const
{
	SourceNext=SourceBeginning;
	DestNext=DestBeginning;

	result Result=ok;
	utf32 Char;
	from_type*Next;
	while(SourceNext<SourceEnd&&
		(Result=fromUTF(SourceNext,SourceEnd,Next,Char))==ok
		)
	{
		Result=toUTF(Char,DestNext,DestEnd,DestNext);
		if(Result!=ok)
			return Result;

		// Conversion was ok. The DestNext field is already updated, so the only thing
		// left is the source buffer pointer.
		SourceNext=Next;
	}

	return Result;
}

/ *
int UTF8CodeConverter::do_length(
	UTF8CodeConverter::state_type&State
	,const UTF8CodeConverter::to_type*Beginning
	,const UTF8CodeConverter::to_type*End
	,size_t Count
	)const throw()
{
	int Result=0;
	while(Beginning<End)
	{
}
virtual int do_length(_St& _State, const _To *_F1,
    const _To *_L1, size_t _N2) const _THROW0()
    {int _Ans = 0;
    for (; _F1 != _L1; )
        if ((*_F1 & 0x80) == 0x00)
            ++_F1, ++_Ans;
        else if ((*_F1 & 0xe0) == 0xc0
            && _F1 + 1 != _L1
            && (_F1[1] & 0xc0) == 0x80)
            _F1 += 2, _Ans += 2;
        else if ((*_F1 & 0xf0) == 0xe0
            && _F1 + 1 != _L1
            && _F1 + 2 != _L1
            && (_F1[1] & 0xc0) == 0x80
            && (_F1[2] & 0xc0) == 0x80)
            _F1 += 3, _Ans += 3;
        else
            break;  // erroneous code
    return (_Ans); }
	* /
*/
END_MATROSKA_NAMESPACE
