/*
	libmatroska : parse and create Matroska files, see http://matroska.sourceforge.net/
	Copyright (C) 2002 Ingo Ralf Blum

	This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

/*!
	\file
	\version \$Id: Writer.hpp,v 1.12 2002/05/04 17:13:44 robux4 Exp $
*/
#ifndef LIBMATROSKA_WRITER_HPP
#define LIBMATROSKA_WRITER_HPP

#include <cassert>
#include <memory>

#include "Config.hpp"
#include "IOCallback.hpp"

START_MATROSKA_NAMESPACE

template<class TYPE,size_t ELEMENT_SIZE,size_t ELEMENT_COUNT> class PackedArray
{
private:
	uint8 Buffer[(ELEMENT_SIZE*ELEMENT_COUNT+7)/8];

public:
	class Reference
	{
	private:
		PackedArray&Array;
		size_t Index;

	public:
		Reference(PackedArray&Array,size_t Index)throw()
			:Array(Array)
			,Index(Index)
		{
			assert(&Array!=0);
		}

		Reference&operator=(TYPE Value)throw()
		{
			Array.set(Index,Value);
			return*this;
		}

		operator TYPE()const throw()
		{
			return Array.get(Index);
		}
	};

	void set(size_t Index,TYPE Value)
	{
		assert(Index<ELEMENT_COUNT);
		assert(Value>=0);
		assert(Value<(1<<ELEMENT_SIZE));

		size_t BitIndex=ELEMENT_SIZE*Index;
		size_t ByteIndex=BitIndex/8;
		BitIndex%=8; // Bit index is from high to low

		// Use ints rather than size_t because the following shift
		// operationg might need a negative offset.
		for(int Available=ELEMENT_SIZE;Available>0;)
		{
			int Copyable=Available;
			if(Copyable>8)
				Copyable=8;
			if(Copyable>(8-BitIndex))
				Copyable=8-BitIndex;

			uint8 Or=Value>>(Available-Copyable-BitIndex);
			uint8 Mask=((1<<Copyable)-1)<<(8-BitIndex-Copyable);
			Buffer[ByteIndex]=Buffer[ByteIndex]&~Mask|Or&Mask;

			Available-=Copyable;
			BitIndex+=Copyable;
			if(BitIndex>=8)
			{
				BitIndex-=8;
				++ByteIndex;
			}
		}
	}

	TYPE get(size_t Index)throw()
	{
		assert(Index<ELEMENT_COUNT);
		
		size_t BitIndex=ELEMENT_SIZE*Index;
		size_t ByteIndex=BitIndex/8;
		BitIndex%=8; // Bit index is from high to low

		TYPE Value=0;
		for(int Available=ELEMENT_SIZE;Available>0;)
		{
			int Copyable=Available;
			if(Copyable>8)
				Copyable=8;
			if(Copyable>(8-BitIndex))
				Copyable=8-BitIndex;

			Value=(Value<<Copyable)
				|(Buffer[ByteIndex]>>(8-BitIndex-Copyable))&((1<<Copyable)-1);
		}

		return Value;
	}

	TYPE operator[](size_t Index)const{return get(Index);}
	Reference operator[](size_t Index){return Reference(*this,Index);}
};

#pragma pack(1)
/*struct MainHeader
{
	ascii TypeHeader[0xa0];
	uint8 Version;
	uint8 MinimumReadVersion;
	uint8 AbsoluteMinimumReadVersion;
	uint8 Flags;
	big_time TotalLength;
	big_uint32 FirstTrackEntryPosition;
	big_uint32 FirstBlockPosition;
	big_uint64 FirstSeekEntryPosition;
	big_uint32 BlockCount;
	uint8 TrackCount;
	uint8 SeekEntrySize;
	uint8 BlockHeaderSize;
	uint8 ClusterHeaderSize;
	uint16 TrackEntrySize;
	uint16 SubtrackEntrySize;
	big_checksum SeekEntryChecksum;
	
	uint8 Reserved0[56];

	utf8 NextPartFilename[0x40];
	big_time NextPartStartTime;

	uint8 Reserved1[3];

	uint8 AttachedFileCount;
	big_uint64 FirstAttackEntryPosition;
	big_uint32 AttachEntrySize;
	big_checksum AttachEntryChecksum;

	big_uint64 ChapterPosition;
	big_uint32 ChapterSize;
	big_checksum ChapterChecksum;
	big_uint16 ChapterBlockSize;
	big_uint16 EditionEntrySize;
	uint8 EditionCount;

	uint8 Reserved2[147];

	utf8 OriginalFilename[64];
	uint64 SignaturePosition;
	uint16 SignatureSize;

	uint8 Reserved3[434];

	big_checksum ActualHeaderChecksum;
};

struct ExtendedInfo
{
	utf8 Title[192];
	utf8 Edition[128];

	utf8 OriginalAuthorName[128];
	ascii OriginalAuthorURL[128];
	ascii OriginalAuthorMail[64];

	utf8 EncoderName[128];
	ascii EncoderURL[128];
	ascii EncoderMail[64];

	utf8 Comments[2048];

	uint8 Rating[6];

	uint8 ContentSubtype;
	uint8 ContentType;

	big_unixtime EncodingTime;
	big_unixtime LastEditingTime;
	big_uint16 ProductionYear;
	ascii Country[2];

	uint8 Language[4];

	uint8 Reserved[24];

	big_checksum Checksum;
};

struct ContentInfo
{
	uint8 Reserved[1020];
	big_checksum Checksum;
};*/
/*struct SeekEntry
{
	big_time Time; // Time of the first keyframe in the cluster
	big_uint64 ClusterPosition;
};

struct AttachmentEntry
{
	utf8 Description[128];
	ascii MIME[32];
	filename Filename;

	big_uint64 DataPosition;
	big_uint32 DataSize;
	big_checksum DataChecksum;

	uint8 Compression;
};

struct EditionEntry
{
	utf8 Name[64];
	big_uint16 ChapterCount;
};

struct ChapterEntry
{
	big_time Beginning,End;
	utf8 Name[256];
};
#pragma pack()*/

/*class MATROSKA_EXPORT Writer
{
private:
	IOCallback*Callback;
	std::auto_ptr<IOCallback> StandardCallback;

public:
	Writer()throw();
	~Writer()throw();

	// Open an existing file.
	void create(const char*Path);
	void create(IOCallback&Callback);

	IOCallback&getCallback()const throw()
	{
		assert(Callback!=0);
		return*Callback;
	}

	bool isOpen()const throw(){return Callback!=0;}

	void close();
};*/

END_MATROSKA_NAMESPACE

#endif
