// Package:
//  This file is part of the Ingo++ library. Contact Ingo Ralf Blum for more information.
//
// Copyright: Copyright Ingo Ralf Blum, 2001
// You are not allowed to remove this header from this file.
//
// Authors:
//  Ingo Ralf Blum <ingoralfblum@users.sf.net>, <papierkorb@ingoralfblum.com>, <ingoralfblum@gmx.de>
// 
// File version: $Revision: 1.3 $
//
// License: LGPL
//  You should have received a copy of the GNU (Lesser) General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

#include "Config.hpp"
#include "StdInclude.hpp"
#include "WindowsFile.hpp"

#include "../../Debug.hpp"

#include <sstream>
#include <memory>

using namespace std;

START_MATROSKA_NAMESPACE

WindowsFile::WindowsFile()
{}

WindowsFile::WindowsFile(const char*Filename
	,DWORD Access
	,DWORD ShareMode
	,LPSECURITY_ATTRIBUTES SecurityAttrs
	,DWORD CreationDist
	,DWORD FlagsAndAttrs
	,HANDLE Template
	)
{
	open(Filename,Access,ShareMode,SecurityAttrs,CreationDist,FlagsAndAttrs,Template);
}

void WindowsFile::open(const char*FileName
	,DWORD Access
	,DWORD ShareMode
	,LPSECURITY_ATTRIBUTES SecurityAttrs
	,DWORD CreationDist
	,DWORD FlagsAndAttrs
	,HANDLE Template
	)
{
	assert(FileName!=0);

	close();

	assert(!isOpen());

	HANDLE Handle=CreateFileA(FileName,Access,ShareMode,SecurityAttrs,CreationDist,FlagsAndAttrs,Template);
	if(Handle==INVALID_HANDLE_VALUE)
	{
		ostringstream Msg;
		Msg<<"Can't open \""<<FileName<<"\"";
		throw WindowsError(Msg.str());
	}

	attach(Handle);
}

void WindowsFile::close()
{
	WindowsHandle::close();
}

size_t WindowsFile::write(const void*Mem,size_t Len)
{
	assert(getHandle()!=INVALID_HANDLE_VALUE);
	assert(Mem!=0);

	DWORD Written=0;
	if(!WriteFile(getHandle(),Mem,Len,&Written,0))
	{
		ostringstream Msg;
		Msg<<"Error while writing "<<Len<<" bytes";
		throw WindowsError(Msg.str());
	}
	
	return Written;
}

uint32 WindowsFile::read(void*Mem,size_t Len)
{
    assert(isOpen());
    assert(Mem!=0);

    DWORD BytesRead=0;
    if(!ReadFile(getHandle(),Mem,Len,&BytesRead,0))
    {
	    ostringstream Msg;
	    Msg<<"Error while reading "<<Len<<" bytes to "<<Mem;
	    throw WindowsError(Msg.str());
    }

    return BytesRead;
    }

void WindowsFile::setFilePointer(int64 Offset,seek_mode Whence)
{
	seek(Offset,Whence);
}

uint64 WindowsFile::getFilePointer()
{
	return seek(0,seek_current);
}

uint64 WindowsFile::seek(int64 Offset,seek_mode RelTo)const
{
	assert(isOpen());

	DWORD Method;
	switch(RelTo)
	{
	case seek_beginning:
		Method=FILE_BEGIN;
		break;
	case seek_current:
		Method=FILE_CURRENT;
		break;
	case seek_end:
		Method=FILE_END;
		break;
	default:
		assert(false);
		break;
	}

	LONG High=Offset>>32;
	DWORD Low=SetFilePointer(getHandle(),Offset&0xffffffff,&High,Method);
	if(Low==INVALID_SET_FILE_POINTER)
	{
		DWORD Error=GetLastError();
		if(Error!=NO_ERROR)
		{
			ostringstream Msg;
			Msg<<"WindowsFile::seek: Can't seek to "<<(int)Offset<<" relative to "<<RelTo; // TODO: The int cast is bad
			throw WindowsError(Error,Msg.str());
		}
	}

	return (((uint64)High)<<32)|Low;
}

DWORD WindowsFile::getAttributes(const char*Filename)
{
	assert(Filename!=0);
	DWORD Attrs=GetFileAttributesA(Filename);
	if(Attrs==-1)
	{
		ostringstream Msg;
		Msg<<"Can't get file attributes for \""<<Filename<<"\"";
		throw WindowsError(Msg.str());
	}

	return Attrs;
}

void WindowsFile::flush()
{
	assert(isOpen());
	if(FlushFileBuffers(getHandle())==FALSE)
		throw WindowsError("Can't flush file buffers");
}

uint64 WindowsFile::getSize()
{
	assert(isOpen());

	DWORD High;
	DWORD Low=GetFileSize(getHandle(),&High);
	if(Low==INVALID_FILE_SIZE)
	{
		DWORD Error=GetLastError();
		if(Error!=NO_ERROR)
			throw WindowsError(Error,"Can't retrieve the file size for a windows file");
	}

	return (uint64)Low|(((uint64)High)<<32);
}

size_t WindowsFile::ioControl(DWORD Operation,const void*InBuffer,size_t InBufferSize,void*OutBuffer,size_t OutBufferSize)const
{
	assert(isOpen());
	assert(InBuffer!=0||InBufferSize==0);
	assert(OutBuffer!=0||OutBuffer==0);

	DWORD NumberOfBytesRead=0;
	if(DeviceIoControl(getHandle(),Operation,const_cast<void*>(InBuffer),InBufferSize,OutBuffer,OutBufferSize,&NumberOfBytesRead,0)==FALSE)
		throw WindowsError("Error in DeviceIoControl");

	assert(NumberOfBytesRead<=OutBufferSize);

	return NumberOfBytesRead;
}

void getWindowsDirectory(std::string&Result)
{
	assert(false); // TODO: Test this function

	size_t Size=GetWindowsDirectoryA(0,0);
	//boost::scoped_array<char> Buffer(new char[Size+1]);
	std::auto_ptr<char> Buffer(new char[Size+1]);
	MATROSKA_ASSERT_NEW(Buffer.get());

	Size=GetWindowsDirectoryA(Buffer.get(),Size+1);
	if(Size==0)
		throw WindowsError("Can't retrieve windows directory");

	Result=Buffer.get();
}

std::string WindowsFile::getShortPath(const char*LongPath)
{
	assert(LongPath!=0);

	char Dummy;
	DWORD Length=GetShortPathNameA(LongPath,&Dummy,1);
	if(Length==0)
	{
		DWORD Error=GetLastError();
		ostringstream Msg;
		Msg<<"Can't determine the short path name length of \""<<LongPath<<"\"";
		throw WindowsError(Error,Msg.str());
	}
	else if(Length==1)
		return std::string(&Dummy,1);

	//boost::scoped_array<char> Buffer(new char[Length+1]);
	std::auto_ptr<char> Buffer(new char[Length+1]);
	MATROSKA_ASSERT_NEW(Buffer.get());
	Length=GetShortPathNameA(LongPath,Buffer.get(),Length+1);
	if(Length==0)
	{
		DWORD Error=GetLastError();
		ostringstream Msg;
		Msg<<"Can't determine the short path name of \""<<LongPath<<"\"";
		throw WindowsError(Error,Msg.str());
	}

	return std::string(Buffer.get(),Length);
}


std::string WindowsFile::getFullPath(const char*Filename)
{
	assert(Filename!=0);

	char Dummy;
	DWORD Length=GetFullPathNameA(Filename,1,&Dummy,0);
	if(Length==0)
	{
		DWORD Error=GetLastError();
		ostringstream Msg;
		Msg<<"Can't determine the short path name length of \""<<Filename<<"\"";
		throw WindowsError(Error,Msg.str());
	}
	else if(Length==1)
		return std::string(&Dummy,1);

	//boost::scoped_array<char> Buffer(new char[Length+1]);
	std::auto_ptr<char> Buffer(new char[Length+1]);
	MATROSKA_ASSERT_NEW(Buffer.get());
	Length=GetFullPathNameA(Filename,Length+1,Buffer.get(),0);
	if(Length==0)
	{
		DWORD Error=GetLastError();
		ostringstream Msg;
		Msg<<"Can't determine the short path name of \""<<Filename<<"\"";
		throw WindowsError(Error,Msg.str());
	}

	return std::string(Buffer.get(),Length);
}

void WindowsFile::getAttributes(const char*Path,WIN32_FILE_ATTRIBUTE_DATA&Data)
{
	assert(Path!=0);
	assert(&Data!=0);

	if(!GetFileAttributesExA(Path,GetFileExInfoStandard,&Data))
		throw WindowsError("Error in GetFileAttributesExA");
}

uint64 WindowsFile::getSize(const char*Path)
{
	WIN32_FILE_ATTRIBUTE_DATA Data;
	getAttributes(Path,Data);
	return ((uint64)Data.nFileSizeHigh<<32)|Data.nFileSizeLow;
}

END_MATROSKA_NAMESPACE
