// Package:
//  This file is part of the Ingo++ library. Contact Ingo Ralf Blum for more information.
//
// Copyright: Copyright Ingo Ralf Blum, 2001
// You are not allowed to remove this header from this file.
//
// Authors:
//  Ingo Ralf Blum <ingoralfblum@users.sf.net>, <papierkorb@ingoralfblum.com>, <ingoralfblum@gmx.de>
// 
// File version: $Revision: 1.2 $
//
// License: LGPL
//  You should have received a copy of the GNU (Lesser) General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

#include "Config.hpp"
#include "StdInclude.hpp"
#include "WindowsHandle.hpp"
#include "WindowsError.hpp"

#include "../../Debug.hpp"

START_MATROSKA_NAMESPACE

WindowsHandle::WindowsHandle()throw()
{
	Handle=INVALID_HANDLE_VALUE;
}

WindowsHandle::WindowsHandle(HANDLE Handle)throw()
{
	this->Handle=Handle;
}

WindowsHandle::~WindowsHandle()
{
	try
	{
		close();
	}
	catch(std::exception&Ex)
	{
		MATROSKA_TRACE("Error in CloseHandle in destructor is ignored: %hs",Ex.what());
	}
}

WindowsHandle::multi_wait_result WindowsHandle::wait(size_t Count,const HANDLE*Handles,bool WaitAll,DWORD Milliseconds)
{
#ifdef _DEBUG
	assert(Count<=MAXIMUM_WAIT_OBJECTS);
	assert(Handles!=0);
	for(size_t i=0;i<Count;++i)
		assert(Handles[i]!=0&&Handles[i]!=INVALID_HANDLE_VALUE);
#endif

	DWORD Result=WaitForMultipleObjects(Count,Handles,WaitAll?TRUE:FALSE,Milliseconds);
	if(Result>=WAIT_OBJECT_0&&Result<WAIT_OBJECT_0+Count)
		return multi_wait_result(wait_succeeded,Result-WAIT_OBJECT_0);
	else if(Result>=WAIT_ABANDONED_0&&Result<WAIT_ABANDONED_0+Count)
		return multi_wait_result(wait_abandoned,Result-WAIT_ABANDONED_0);
	else if(Result==WAIT_TIMEOUT)
		return multi_wait_result(wait_timed_out,0);
	else
	{
		if(Result!=WAIT_FAILED)
			assert(false);
		throw WindowsError("Multi wait failed");
	}
}

WindowsHandle::multi_wait_result WindowsHandle::wait(const WindowsHandle&First,const WindowsHandle&Second,DWORD Milliseconds,bool WaitAll)
{
	assert(&First!=0);
	assert(&Second!=0);

	HANDLE Handles[2]=
	{
		First.getHandle()
		,Second.getHandle()
	};

	return wait(sizeof(Handles)/sizeof(Handles[0]),Handles,WaitAll,Milliseconds);
}

bool WindowsHandle::isOpen()const throw()
{
	return Handle!=INVALID_HANDLE_VALUE&&Handle!=0;
}

void WindowsHandle::close()
{
	if(isOpen())
	{
		if(!CloseHandle(Handle))
			throw WindowsError("Error while closing handle");

		Handle=INVALID_HANDLE_VALUE;
	}
}

void WindowsHandle::attach(HANDLE Handle)
{
	close();
	this->Handle=Handle;
}

bool WindowsHandle::isSignaled()const 
{
	return wait(0)==wait_succeeded;
}

WindowsHandle::wait_result WindowsHandle::wait(uint32 Milliseconds)const 
{
	assert(Handle!=0&&Handle!=INVALID_HANDLE_VALUE);

	DWORD Result=WaitForSingleObject(Handle,Milliseconds);
	switch(Result)
	{
	case WAIT_TIMEOUT:
		return wait_timed_out;
	case WAIT_OBJECT_0:
		return wait_succeeded;
	case WAIT_ABANDONED:
		// This should not happen, if the handle is released before the owning threads exists.
		return wait_abandoned;
	default:
		throw WindowsError("Error while retrieving handle state");
	}
}

HANDLE WindowsHandle::getHandle()const throw()
{
	return Handle;
}

WindowsHandle&WindowsHandle::operator=(HANDLE Handle)
{
	attach(Handle);
	return*this;
}

WindowsHandle::operator HANDLE()const throw()
{
	return getHandle();
}

END_MATROSKA_NAMESPACE
