// Package:
//  This file is part of the Ingo++ library. Contact Ingo Ralf Blum for more information.
//
// Copyright: Copyright Ingo Ralf Blum, 2001
//
// Authors:
//  Ingo Ralf Blum <ingoralfblum@users.sf.net>, <papierkorb@ingoralfblum.com>, <ingoralfblum@gmx.de>
// 
// File version: $Revision: 1.2 $
//
// License: LGPL
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

#include "Config.hpp"
#include "StdInclude.hpp"

#include "WindowsModule.hpp"
#include "WindowsError.hpp"
#include "LocalMemory.hpp"

#include "../../Types.hpp"
#include "../../Debug.hpp"

#include <sstream>

using namespace std;

START_MATROSKA_NAMESPACE

HMODULE Instance=0;

void WindowsModule::initialize()
{
	Loaded=false;
	Handle=0;
}

WindowsModule::WindowsModule(HMODULE Handle)throw()
{
	initialize();
	this->Handle=Handle;
}

WindowsModule::WindowsModule(const char*Name,bool Load)
{
	initialize();

	if(Load)
		load(Name);
	else
		query(Name);
}

void WindowsModule::load(const char*Name)
{
	assert(Name!=0);

	close();

	Handle=LoadLibraryA(Name);
	if(Handle==0)
		throw WindowsError(std::string("Can't load library \"")+Name+"\"");
	Loaded=true;
}

WindowsModule::WindowsModule(const char*Name,DWORD Flags)
{
	initialize();
	load(Name,Flags);
}

void WindowsModule::query(const char*Name)
{
	close();
	assert(!Loaded); // Should be set to false by free()

	Handle=GetModuleHandleA(Name);
	if(Handle==0)
	{
		ostringstream Msg;
		Msg<<"Can't get module handle for \""<<Name<<"\"";
		throw WindowsError(Msg.str());
	}
}

void WindowsModule::load(const char*Name,DWORD Flags)
{
	assert(Name!=0);

	close();

	Handle=LoadLibraryExA(Name,0,Flags);
	if(Handle==0)
	{
		ostringstream Msg;
		Msg<<"Can't load library \""<<Name<<"\" with flags 0x"<<hex<<Flags;
		throw WindowsError(Msg.str());
	}
	Loaded=true;
}

WindowsModule::~WindowsModule()throw()
{
	try
	{
		close();
	}
	catch(std::exception&Ex)
	{
		MATROSKA_TRACE("Warning: Error in WindowsModule::close() ignored: %hs\n",Ex.what());
	}
}

void WindowsModule::close()
{
	if(Handle==0)
	{
		assert(!Loaded);
		return;
	}

	if(Loaded)
	{
		if(FreeLibrary(Handle)==FALSE)
			throw WindowsError("Can't free library");
		
		Loaded=false;
	}

	Handle=0;
}

void WindowsModule::disableThreadCalls()const 
{
	assert(Handle!=0);
	if(DisableThreadLibraryCalls(Handle)==FALSE)
		throw WindowsError("Can't disable thread calls for library");
}

FARPROC WindowsModule::getProcAddress(const char*Name,bool ReturnNullIfNotFound)const 
{
	assert(Name!=0);
	assert(Handle!=0);

	FARPROC Proc=GetProcAddress(Handle,Name);
	if(Proc==0)
	{
		DWORD Error=GetLastError();
		if(Error!=ERROR_PROC_NOT_FOUND||!ReturnNullIfNotFound)
			throw WindowsError(Error,std::string("Can't get proc address for \"")+Name+"\"");
	}

	return Proc;
}

std::string WindowsModule::getFilenameA()
{
	char Buffer[MAX_PATH+1];

	// TODO: Avoid overflow of the buffer and expand it if required.
	DWORD Result=GetModuleFileNameA(Handle,Buffer,countof(Buffer));
	if(Result==0)
		throw WindowsError("Error in retrieving module file name");

	return std::string(Buffer,Result);
}

HRSRC WindowsModule::findResource(const char*Type,int ID)const 
{
	HRSRC ResourceHandle=FindResourceA(Handle,MAKEINTRESOURCEA(ID),"BINARY");
	if(ResourceHandle==0)
	{
		ostringstream Msg;
		Msg<<"Unable to find resource with type ";
		if(IS_INTRESOURCE(Type))
			Msg<<reinterpret_cast<int>(Type);
		else
			Msg<<Type;
		Msg<<" and id "<<ID;
		throw WindowsError(Msg.str());
	}

	return ResourceHandle;
}

void*WindowsModule::getResourcePointer(HRSRC Resource)const 
{
	assert(Resource!=0);

	HGLOBAL GlobalHandle=LoadResource(Handle,Resource);
	if(GlobalHandle==0)
		throw WindowsError("Can't load resource");

	void*Pointer=LockResource(GlobalHandle);
	if(Pointer==0)
		throw WindowsError("Can't lock resource");

	return Pointer;
}

size_t WindowsModule::getResourceSize(HRSRC Resource)const 
{
	DWORD Size=SizeofResource(Handle,Resource);
	if(Size==0)
		throw WindowsError("Can't retrieve resource size");

	return Size;
}

std::string WindowsModule::getMessageA(DWORD MessageID,DWORD LanguageID)
{
	char*Buffer=0;
	DWORD Length=FormatMessageA(
		FORMAT_MESSAGE_ALLOCATE_BUFFER|FORMAT_MESSAGE_FROM_HMODULE|FORMAT_MESSAGE_IGNORE_INSERTS
		,Handle
		,MessageID
		,LanguageID
		,reinterpret_cast<char*>(&Buffer)
		,0 // Size not needed
		,0 // No arguments
		);
	if(Length==0)
	{
		ostringstream Msg;
		Msg<<"Can't format message id "<<hex<<MessageID<<" with languag "<<hex<<LanguageID;
		throw WindowsError(Msg.str());
	}

	LocalMemory Mem(reinterpret_cast<HGLOBAL>(Buffer));
	LocalMemory::Flags Flags=Mem.getFlags();
	return reinterpret_cast<const char*>(Mem.lock());
}

END_MATROSKA_NAMESPACE
