/*
	libmatroska : parse and create Matroska files, see http://matroska.sf.net/
	Copyright (C) 2002 Steve Lhomme

	This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

/*!
    \file
    \version \$Id: test8.cpp,v 1.13 2002/05/24 15:53:41 robux4 Exp $
    \brief Test reading of the Header
    \author Lasse Krkkinen <tronic @ users.sf.net>
    \author Steve Lhomme     <robux4 @ users.sf.net>
*/

#include <iostream>

#include "FileKax.hpp"
#include "StdIOCallback.hpp"

using namespace LIBMATROSKA_NAMESPACE;
using namespace std;

void displayTrackInfo(const TrackInfo & aTrackInfo)
{
    cout << "Track name  : " << aTrackInfo.Name << endl;
    cout << "enabled     : " << (aTrackInfo.Enabled?"yes":"no") << endl;
    cout << "preferred   : " << (aTrackInfo.Preferred?"yes":"no") << endl << endl;
}

void displayAudioInfo(const TrackInfoAudio & aAudioTrack)
{
    cout << "sample rate : " << (int)aAudioTrack.SampleRate << endl;
    cout << "bit depth   : " << (int)aAudioTrack.BitDepth << endl;
    cout << "channel/tra : " << (int)aAudioTrack.ChannelPerSubtrack << endl;
    cout << "channel pos : " << (int)aAudioTrack.ChannelPosition << endl;
    cout << "DTS         : " << (aAudioTrack.DTS?"yes":"no") << endl;
    cout << "secondary   : " << (aAudioTrack.SecondaryTrack?"yes":"no") << endl;
    cout << "subtrack #  : " << (int)aAudioTrack.SubtrackNumber << endl;
    cout << "AVI format  : " << (int)aAudioTrack.AVIFormatTag << endl << endl;
}

void displayVideoInfo(const TrackInfoVideo & aVideoTrack)
{
    cout << "interlaced  : " << (aVideoTrack.Interlaced?"yes":"no") << endl;
    cout << "stereo/3D   : " << (aVideoTrack.Stereo3D?"yes":"no");
    if (aVideoTrack.Stereo3D)
	switch (aVideoTrack.Eyes) {
	case eye_right:
	    cout << " (right eye)";
	    break;
	case eye_left:
	    cout << " (left eye)";
	    break;
	case eye_both:
	    cout << " (both eyes)";
	    break;
	default:
	    cout << " (unknown eyes)";
	    break;
	}
    cout << endl;
    cout << "resolution  : " << int(aVideoTrack.xSize) << "x" << int(aVideoTrack.ySize);
    cout << " (aspect ratio : " << (float(aVideoTrack.AspectRatio) / 1000.0) << ")" << endl;
    cout << "FourCC      : " << char(aVideoTrack.FourCC[0]) << char(aVideoTrack.FourCC[1]) << char(aVideoTrack.FourCC[2]) << char(aVideoTrack.FourCC[3]) << endl;
}

/*!
    \note you can generate the file used in this example using test6.cpp
*/
int main(void)
{
    try {
	StdIOCallback Cluster_file("muxed.mcf", MODE_READ);
    
	// create the Matroska file
	FileKax MuxedFile(Cluster_file);

	MuxedFile.ReadHead();
	MuxedFile.ReadTracks();
	MuxedFile.ReadCodec();
	uint8 TrackNumber = MuxedFile.GetTrackNumber();

	TrackInfo *Tracks = new TrackInfo[TrackNumber];
	TrackInfoAudio aAudioTrack;
	TrackInfoVideo aVideoTrack;

	Track * track1 = MuxedFile.GetTrack(1);
	Track * track2 = MuxedFile.GetTrack(2);
	Track * track3 = MuxedFile.GetTrack(3); // should be 0
	// get information about the 1st track

	MuxedFile.Track_GetInfo(track1, Tracks[0]);
	displayTrackInfo(Tracks[0]);

	if (Tracks[0].TrackType == track_audio) {
	    MuxedFile.Track_GetInfo_Audio(track1, aAudioTrack);
	    displayAudioInfo(aAudioTrack);
	}
	else
	    cout << "the expected audio track is not an audio one :(" << endl;

	MuxedFile.Track_GetInfo(track3, Tracks[2]);
	displayTrackInfo(Tracks[2]);

	if (Tracks[2].TrackType == track_video) {
	    MuxedFile.Track_GetInfo_Video(track3, aVideoTrack);
	    displayVideoInfo(aVideoTrack);
	}
	else
	    cout << "the expected video track is not an video one :(" << endl;

	// let's read only track1 (audio)
	MuxedFile.SelectReadingTrack(track1);
	MuxedFile.SelectReadingTrack(track2);
	MuxedFile.SelectReadingTrack(track1,false);
	MuxedFile.SelectReadingTrack(track1);
	MuxedFile.SelectReadingTrack(track2,false);
	MuxedFile.SelectReadingTrack(track2);
	MuxedFile.SelectReadingTrack(track3);

	// read the frames from the selected tracks and put them in an output file
	StdIOCallback Output_file1("out-binr.bin", MODE_CREATE);
	StdIOCallback Output_file2("out-text.bin", MODE_CREATE);
	StdIOCallback Output_file3("out-vide.bin", MODE_CREATE);

	Track * TrackRead;
	uint32 timecode; // not used yet
	binary *aFrame;
	uint32 aFrameSize;
	while (MuxedFile.ReadFrame(TrackRead, timecode, aFrame, aFrameSize))
	{
	    if (TrackRead == track1)
	    {
		Output_file1.write(aFrame, aFrameSize);
	    }
	    else if (TrackRead == track2)
	    {
		Output_file2.write(aFrame, aFrameSize);
	    }
	    else if (TrackRead == track3)
	    {
		Output_file3.write(aFrame, aFrameSize);
	    }
	    else cout << "received a frame from an unwanted track" << endl;
	}
    }
    catch (exception & Ex)
    {
	cout << Ex.what() << endl;
	return -1;
    }

    return 0;
}
