/*
	libmatroska : parse and create Matroska files, see http://matroska.sf.net/
	Copyright (C) 2002 Steve Lhomme

	This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

/*!
    \file
    \version \$Id: test6.cpp,v 1.22 2002/05/23 16:47:13 robux4 Exp $
    \brief Test muxing two tracks into valid clusters/blocks/frames
    \author Lasse Krkkinen <tronic @ users.sf.net>
    \author Steve Lhomme     <robux4 @ users.sf.net>
*/

#ifdef _MSC_VER
#include <windows.h>      // for min/max
#endif // _MSC_VER

#include <iostream>

#include "StdIOCallback.hpp"
#include "FileKax.hpp"
#include "TrackAudio.hpp"
#include "TrackSubtitle.hpp"
#include "TrackVideo.hpp"

using namespace LIBMATROSKA_NAMESPACE;
using namespace std;

unsigned int BIN_FILE_SIZE = 15000;
unsigned int TXT_FILE_SIZE = 3000;
const unsigned int BIN_FRAME_SIZE = 1500;
const unsigned int TXT_FRAME_SIZE = 200;

/*!
    The first file is a "binary" file with data scaling from 0x00 to 0xFF repeatedly
    The second file is a "text" file with data scaling from 'z' to 'a'
*/
int main(int argc, char **argv)
{
    char *file1;
    char *file2;
    if (argc == 1)
    {
	file1 = "binr.bin";
	// create the files
	StdIOCallback bin_file(file1, MODE_CREATE);
	char c;
	uint64 i;
	for (i=0; i<BIN_FILE_SIZE; i++)
	{
	    c = char(i % 0x100);
	    bin_file.write(&c, 1);
	}
	bin_file.close();

	file2 = "text.bin";
	StdIOCallback txt_file(file2, MODE_CREATE);
	for (i=TXT_FILE_SIZE; i>0; i--)
	{
	    c = char((i % 26) + 'a');
	    txt_file.write(&c, 1);
	}
	txt_file.close();
    }
    else if(argc < 3)
    {
	cout << "Syntax: mcfmux [binfile txtfile]" << endl;
	return 1; // error
    }
    else
    {
	file1 = argv[1];
	file2 = argv[2];
    }

    cout << "Creating \"muxed.mcf\"" << endl;

    try {
	// write the head of the file (with everything already configured)
        StdIOCallback out_file("muxed.mcf", MODE_CREATE);

	// create the Matroska file
	FileKas MuxedFile(out_file);

	// add the binary track (audio)
	TrackAudio AudioTrack;
	Track * binID = MuxedFile.CreateTrack( AudioTrack );
	MuxedFile.track_SetName(binID, "Audio test");
	MuxedFile.track_SetLaced(binID);

	TrackInfoAudio aAudioTrack;
	memset(&aAudioTrack, 0, sizeof(aAudioTrack));
	aAudioTrack.SampleRate = 44100;
	aAudioTrack.BitDepth = 8;
	aAudioTrack.ChannelPerSubtrack = 1;
	aAudioTrack.DTS = false;
	aAudioTrack.SecondaryTrack = false;
	aAudioTrack.SubtrackNumber = 1;
	aAudioTrack.ChannelPosition = SPEAKER_FRONT_LEFT | SPEAKER_FRONT_RIGHT;

	MuxedFile.Track_SetInfo_Audio(binID, aAudioTrack);

	// add the text track (subtitle)
	TrackSubtitle SubtitleTrack;
	Track * txtID = MuxedFile.CreateTrack( SubtitleTrack );
	MuxedFile.track_SetName(txtID, "Subtitle test");

	// try a video track
	TrackVideo VideoTrack;
	Track * videoID = MuxedFile.CreateTrack( VideoTrack );
	MuxedFile.track_SetName(videoID, "Video test");

	TrackInfoVideo aVideoTrack;
	memset(&aVideoTrack, 0, sizeof(aVideoTrack));

	aVideoTrack.AspectRatio = 1333; // 4:3
	aVideoTrack.Eyes = eye_both;
	aVideoTrack.FourCC[0] = 'X';
	aVideoTrack.FourCC[1] = 'v';
	aVideoTrack.FourCC[2] = 'i';
	aVideoTrack.FourCC[3] = 'D';
	aVideoTrack.Interlaced = false;
	aVideoTrack.xSize = 640;
	aVideoTrack.ySize = 480;

	MuxedFile.Track_SetInfo_Video(videoID, aVideoTrack);

	// first, write the header in the file
	MuxedFile.RenderHead("libmatroska test6.cpp");

	// to verify the muxing is working
	MuxedFile.SetMaxClusterSize(2000);

	// start muxing (2 binary frames for each text frame)
	StdIOCallback bin_file(file1, MODE_READ);
	StdIOCallback txt_file(file2, MODE_READ);
	
	bin_file.setFilePointer(0L, seek_end);
	txt_file.setFilePointer(0L, seek_end);
	
	BIN_FILE_SIZE = unsigned int(bin_file.getFilePointer());
	TXT_FILE_SIZE = unsigned int(txt_file.getFilePointer());
	
	binary *buf_bin = new binary[BIN_FILE_SIZE];
	binary *buf_txt = new binary[TXT_FILE_SIZE];

	bin_file.setFilePointer(0L, seek_beginning);
	txt_file.setFilePointer(0L, seek_beginning);

	bin_file.read(buf_bin, BIN_FILE_SIZE);
	txt_file.read(buf_txt, TXT_FILE_SIZE);

	binary *curs_bin = buf_bin;
	binary *curs_txt = buf_txt;
	unsigned int remain_bin = BIN_FILE_SIZE;
	unsigned int remain_txt = TXT_FILE_SIZE;

	while (remain_bin !=0 || remain_txt != 0)
	{
	    // a text chunck
	    if (remain_txt !=0) {
		MuxedFile.AddFrame(txtID, 0, curs_txt, min(TXT_FRAME_SIZE,remain_txt));
		curs_txt += min(TXT_FRAME_SIZE,remain_txt);
		remain_txt -= min(TXT_FRAME_SIZE,remain_txt);
	    }
	    // a binary chunck
	    if (remain_bin !=0) {
//		MuxedFile.AddFrame(binID, 0, curs_bin, min(BIN_FRAME_SIZE,remain_bin));
		MuxedFile.AddFrame(videoID, 0, curs_bin, min(BIN_FRAME_SIZE,remain_bin));
		curs_bin += min(BIN_FRAME_SIZE,remain_bin);
		remain_bin -= min(BIN_FRAME_SIZE,remain_bin);
	    }
	    // a text chunck
	    if (remain_txt !=0) {
		MuxedFile.AddFrame(txtID, 0, curs_txt, min(TXT_FRAME_SIZE,remain_txt));
		curs_txt += min(TXT_FRAME_SIZE,remain_txt);
		remain_txt -= min(TXT_FRAME_SIZE,remain_txt);
	    }
	}

	delete[] buf_bin;
	delete[] buf_txt;

	MuxedFile.Close(1000); // 1000 ms
    }
    catch (exception & Ex)
    {
	cout << Ex.what() << endl;
    }

    return 0;
}
