/**
	\file wav2kas.cpp
	Convert RIFF/WAV files into the corresponding Matroska files
*/

#include "StdIOCallback.hpp"
#include <FileKax.hpp>
#include <TrackAudio.hpp>
#include <stdio.h>

#include "kaswav.hpp"

using namespace LIBMATROSKA_NAMESPACE;

const char version[] = "0.1.0";

void usage()
{
    printf("This program converts a RIFF/WAV file into an Matroska file\n\n");
    printf("Usage: wav2kas <input-filename> [output-filename]\n");
    printf("\tif no output file is specified the input filename is appended with .kas\n");
}

int main(int argc,char *argv[])
{
    char app_name[250];
    sprintf(app_name, "wav2kas v%s",version);
    printf("wav2kas v%s - (c) 2002 Steve Lhomme - GPL License\n", version);
    // check for the valid number of parameters
    if (argc != 2 && argc != 3)
    {
	usage();
	return -1;
    }

    // open the input file
    std::string input_Filename = argv[1];
    WAV_file input_File(input_Filename);
    
    if (!input_File.Open(RIFF_file::RIFF_READ))
    {
	printf("Impossible to read the file '%s'\n", input_Filename.c_str());
	return -1;
    }

    // open the output file
    std::string output_Filename;
    if (argc == 3)
        output_Filename = argv[2];
    else
        output_Filename = input_Filename + ".kas";
    
    StdIOCallback output_File(output_Filename.c_str(), MODE_CREATE);

    input_File.ValidateHeader();

    FMT_chunk * fmtChunk = NULL;
    DATA_chunk * dataChunk = NULL;
    BEXT_chunk * bextChunk = NULL;
    LIST_chunk * listChunk = NULL;
    LIST_Info * listInfo = NULL;
    while (!dataChunk && input_File.FindNextChunk())
    {
	switch(input_File.ChunkType())
	{
	case WAV_file::DATA_CHUNK:
	    dataChunk = new DATA_chunk(input_File.AcquireChunk());
	    input_File.SetDataChunk(dataChunk);
	    break;
	case WAV_file::FORMAT_CHUNK:
	    fmtChunk = new FMT_chunk(input_File.AcquireChunk());
	    input_File.SetFormatChunk(fmtChunk);
	    fmtChunk->Read(input_File.IOfile());
	    break;
	case WAV_file::BROADCAST_CHUNK:
	    bextChunk = new BEXT_chunk(input_File.AcquireChunk());
	    bextChunk->Read(input_File.IOfile());
	    break;
	case WAV_file::LIST_CHUNK:
	    listChunk = new LIST_chunk(input_File.AcquireChunk());
	    switch(listChunk->FindType(input_File.IOfile()))
	    {
	    case LIST_chunk::LIST_INFO:
		listInfo = new LIST_Info;
		if (listInfo == NULL)
		{
		    printf("No more space to save LIST INFO\n");
		    listChunk->SkipData(input_File.IOfile());
		}
		else
		{
		    if (!listInfo->Read(input_File.IOfile()))
		    {
			listChunk->SkipData(input_File.IOfile());
			delete listInfo;
			listInfo = NULL;
		    }
		}
		break;
	    default:
		listChunk->SkipData(input_File.IOfile());
		break;
	    }
	    delete listChunk;
	    listChunk = NULL;
	    break;
	default:
	    // unknown chunk, skip data
	    RIFF_chunk * Unknown = input_File.AcquireChunk();
	    Unknown->SkipData(input_File.IOfile());
	    break;
	}
    }

    if (!dataChunk)
    {
	printf("Impossible to find the data part in '%s'\n", input_Filename.c_str());
	return -1;
    }

    if (!fmtChunk)
    {
	printf("Impossible to find the format part in '%s'\n", input_Filename.c_str());
	return -1;
    }

    FileKas McfFile(output_File);
    TrackAudio AudioTrack;
    Track * aTrack = McfFile.CreateTrack( AudioTrack );

    McfFile.track_SetName(aTrack, "wav2kas generated"); /// \todo should be optional
    McfFile.track_SetLaced(aTrack); /// \todo should be optional

    TrackInfoAudio aAudioTrack;
    memset(&aAudioTrack, 0, sizeof(aAudioTrack));
    aAudioTrack.SampleRate = fmtChunk->SampleRate();
    aAudioTrack.BitDepth = fmtChunk->BitDepth();
    aAudioTrack.ChannelPerSubtrack = fmtChunk->ChannelPerSubtrack();
    aAudioTrack.DTS = false;
    aAudioTrack.SecondaryTrack = false;
    aAudioTrack.SubtrackNumber = 1;
    aAudioTrack.ChannelPosition = SPEAKER_FRONT_LEFT | SPEAKER_FRONT_RIGHT;

    McfFile.Track_SetInfo_Audio(aTrack, aAudioTrack);

    /// \todo put the codec header data here, including the data type (AVI compatible)

    // first, write the header in the file
    McfFile.RenderHead(app_name);

    // read data from the input file and put them in the output file
    const uint32 maxRead = 2000;
    binary DataBuffer[maxRead];
    uint32 timecode = input_File.GetCurrentTimecode();
    uint32 readSize = input_File.ReadData(DataBuffer, maxRead);
    while (readSize != 0)
    {
	McfFile.AddFrame(aTrack, timecode, DataBuffer, readSize);
	timecode = input_File.GetCurrentTimecode();
	readSize = input_File.ReadData(DataBuffer, maxRead);
    }

timecode = input_File.GetCurrentTimecode();

    McfFile.Close(timecode);

    if (fmtChunk != NULL)
	delete fmtChunk;

    if (dataChunk != NULL)
	delete dataChunk;

    if (bextChunk != NULL)
	delete bextChunk;

    if (listChunk != NULL)
	delete listChunk;

    if (listInfo != NULL)
	delete listInfo;

    return 0;
}
