/*
    This file is part of mpa2kas.

    Mpa2kas is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    Mpa2kas is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Foobar; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
/**
    \file mpa2kas.cpp
    Convert MPEG I/II Layer 1/2/3 to Matroska files
    \todo remove bad frames
    \todo support for Layer 1 and 2
    \todo the ID3v1 and v2 data could be put in a the info data
*/

#include "StdIOCallback.hpp"
#include <FileKas.hpp>
#include <TrackAudio.hpp>
#include <stdio.h>
#include "mpafile.h"

using namespace LIBMATROSKA_NAMESPACE;

const char version[] = "0.1.0";

void usage()
{
    printf("This program converts a MPEG-I/II Layer 1,2,3 file into an Matroska file\r\n\r\n");
    printf("Usage: mpa2kas <input-filename> [output-filename]\n");
    printf("\tif no output file is specified the input filename is appended with .kas\r\n");
}

int main(int argc,char *argv[])
{
    bool bVerbose = true;
    char app_name[250];
    sprintf(app_name, "mpa2kas v%s",version);
    printf("wav2kas v%s - (c) 2002 Steve Lhomme - GPL License\n", version);
    // check for the valid number of parameters
    if (argc != 2 && argc != 3)
    {
	usage();
	return -1;
    }

    // open the input file
    std::string input_Filename = argv[1];

    MPA_File input_File(input_Filename, MODE_READ);

    // open the output file
    std::string output_Filename;
    if (argc == 3)
        output_Filename = argv[2];
    else
        output_Filename = input_Filename + ".kas";
    
    StdIOCallback output_File(output_Filename.c_str(), MODE_CREATE);

    // prepare to read the whole file for statistics
    if (bVerbose)
	printf("Reading statistics in %s\r\n", input_Filename.c_str());

    input_File.SkipToFrameStart();

    MPA_Lame_Frame infoFrame;
    bool bInfoFrameOn = false;
    if (input_File.CheckInfoFrame())
    {
	*static_cast<MPA_Frame *>(&infoFrame) = input_File.GetInfoFrame();
	bInfoFrameOn = true;
    }

    uint32 FrameNumber = 0;
    uint64 SampleNumbers = 0;
    uint64 TotalFrameSize = 0;
    uint16 SamplingRate = 0;
    uint16 Channels = 0;
    uint16 SampleSize = 0;
    MPA_Frame frame;
    
    while (input_File.FindNextSyncWord())
    {
	FrameNumber++;
	frame = input_File.GetFrame();
	/// \todo handle the bitrate of the frame in a key/value table, so that we can detect real CBR files
	TotalFrameSize += frame.DataSize() + 4;
	SampleNumbers += frame.SampleSize();
	SamplingRate = frame.SamplingRate();
	Channels = frame.Channels();
	SampleSize = frame.SampleSize();

	input_File.SkipData();
    }

    uint64 totalByteSize = TotalFrameSize * SamplingRate;
    uint32 AverageBitrate, AverageByterate;
    if ((8 * totalByteSize) % SampleNumbers > SampleNumbers>>1)
	AverageBitrate = 1 + 8 * totalByteSize / SampleNumbers;
    else
	AverageBitrate = 8 * totalByteSize / SampleNumbers;

    if (AverageBitrate % 8 > 4)
	AverageByterate = 1 + (AverageBitrate >> 3);
    else
	AverageByterate = AverageBitrate >> 3;

    // write output file
    FileKas McfFile(output_File);
    TrackAudio AudioTrack;
    Track * aTrack = McfFile.CreateTrack( AudioTrack );

    McfFile.track_SetName(aTrack, "mpa2kas generated"); /// \todo should be optional
//    McfFile.track_SetLaced(aTrack); /// \todo should be optional

    TrackInfoAudio aAudioTrack;
    memset(&aAudioTrack, 0, sizeof(aAudioTrack));
    aAudioTrack.SampleRate = SamplingRate;
    aAudioTrack.BitDepth = 16;
    aAudioTrack.ChannelPerSubtrack = Channels;
    aAudioTrack.DTS = false;
    aAudioTrack.SecondaryTrack = false;
    aAudioTrack.SubtrackNumber = 1;
    aAudioTrack.ChannelPosition = SPEAKER_FRONT_LEFT | SPEAKER_FRONT_RIGHT;

    McfFile.Track_SetInfo_Audio(aTrack, aAudioTrack);

    McfFile.RenderHead(app_name);

    /// \todo put the codec header data here, including the data type (AVI compatible)

    // first, write the header in the file
    // prepare to read for output
    input_File.SkipToFrameStart();

    /// \todo add the option to skip the Info frame or not
    if (bInfoFrameOn)
    {
	input_File.FindNextSyncWord();
	input_File.GetFrame();
	input_File.SkipData();
    }
    
    if (bVerbose)
	printf("Writing frames in %s\r\n", output_Filename.c_str());

    binary frame_buffer[5000+4]; // should be enough for 32kHz/320kbps
    uint32 prevSamplesNumber = 0;
    SampleNumbers = 0;
    int i=0;
    while (input_File.FindNextSyncWord())
    {
	frame = input_File.GetFrame();
	SampleNumbers += frame.SampleSize();
	memcpy(frame_buffer, (const binary *)(frame), 4);
	input_File.GetData(&frame_buffer[4], 5000);
	/// \todo handle timecode
    if (bVerbose) {
printf("frame %4d : samples %d ", i++, SampleNumbers);
printf("in pos %d ", input_File.getFilePointer());
printf("size %d\r\n", frame.DataSize()+4);
	}
	McfFile.AddFrame(aTrack, (prevSamplesNumber * 1000) / SamplingRate , frame_buffer, frame.DataSize()+4);
	prevSamplesNumber = SampleNumbers;
    }

    McfFile.Close((SampleNumbers * 1000) / SamplingRate);

    output_File.close();
    input_File.close();

    return 0;
}
